"""
Link Aggregation Control Protocol
IEEE 802.3ad
"""
from packetracer import packetracer, triggerlist
from packetracer.structcbs import unpack_BB

# TLVs length in bytes
LACP_HEADER_LEN = 2
LACP_ACTOR_LEN = 20
LACP_PARTNER_LEN = 20
LACP_COLLECTOR_LEN = 16
LACP_TERMINATOR_LEN = 2
LACP_RESERVED_LEN = 50
# TLV types
ACTOR_TYPE = 1
PARTNER_TYPE = 2
COLLECTOR_TYPE = 3
TERMINATOR_TYPE = 0
DEFAULT_PRIORITY = 0x8000

# Bit encoding of the Actor_State and Partner_State fields
STATE_FIELD_BIT_ORDERING = {
    "expired": (0x80, 7),
    "defaulted": (0x40, 6),
    "distribute": (0x20, 5),
    "collect": (0x10, 4),
    "synch": (0x08, 3),
    "aggregate": (0x04, 2),
    "timeout": (0x02, 1),
    "activity": (0x01, 0),
}


def get_property_of_state_field(varname):
    """Create a get/set-property for bits encoding of state field in Actor and Partner TLVs."""
    mask = STATE_FIELD_BIT_ORDERING.get(varname)[0]
    bit_order = STATE_FIELD_BIT_ORDERING.get(varname)[1]
    return property(
        lambda obj: (obj.state & mask) >> bit_order,
        lambda obj, val: obj.__setattr__("state",
                                         (obj.state & ~mask) | (val << bit_order)),
    )


class LACP(packetracer.Packet):
    __hdr__ = (
        ("subtype", "B", 1),
        ("version", "B", 1),
        ("tlvlist", None, triggerlist.TriggerList),
    )

    def _dissect(self, buf):
        self._init_triggerlist("tlvlist", buf[LACP_HEADER_LEN:], self.__parse_tlv)
        return len(buf)

    @staticmethod
    def __parse_tlv(buf):
        """Parse LACP TLVs and return them as list."""
        tlvlist = []
        shift = 0
        tlv_type, tlv_len = 1, 1
        while (tlv_type | tlv_len) != 0:
            tlv_type, tlv_len = unpack_BB(buf[shift:shift + 2])
            clz = LACP_TLV_CLS.get(tlv_type)
            if not clz == LACP_TLV_CLS.get(0):
                tlv_body = buf[shift: tlv_len + shift]
                shift += tlv_len
            else:
                tlv_body = buf[shift: LACP_TERMINATOR_LEN + shift]
                shift += LACP_TERMINATOR_LEN
            tlvlist.append(clz(tlv_body))
        tlvlist.append(LACPReserved(buf[shift: shift + LACP_RESERVED_LEN]))
        return tlvlist


class LACPActorInfoTlv(packetracer.Packet):
    __hdr__ = (
        ("type", "B", ACTOR_TYPE),
        ("len", "B", LACP_ACTOR_LEN),
        ("sysprio", "H", DEFAULT_PRIORITY),
        ("sys", "6s", b"\x00" * 6),
        ("key", "H", 0),
        ("portprio", "H", DEFAULT_PRIORITY),
        ("port", "H", 1),
        ("state", "B", 0),
        ("reserved", "3s", b"\x00" * 3),
    )

    sys_s = packetracer.get_property_mac("sys")
    expired = get_property_of_state_field("expired")
    defaulted = get_property_of_state_field("defaulted")
    distribute = get_property_of_state_field("distribute")
    collect = get_property_of_state_field("collect")
    synch = get_property_of_state_field("synch")
    aggregate = get_property_of_state_field("aggregate")
    timeout = get_property_of_state_field("timeout")
    activity = get_property_of_state_field("activity")


class LACPPartnerInfoTlv(packetracer.Packet):
    __hdr__ = (
        ("type", "B", PARTNER_TYPE),
        ("len", "B", LACP_PARTNER_LEN),
        ("sysprio", "H", DEFAULT_PRIORITY),
        ("sys", "6s", b"\x00" * 6),
        ("key", "H", 0),
        ("portprio", "H", DEFAULT_PRIORITY),
        ("port", "H", 1),
        ("state", "B", 0),
        ("reserved", "3s", b"\x00" * 3),
    )

    sys_s = packetracer.get_property_mac("sys")
    expired = get_property_of_state_field("expired")
    defaulted = get_property_of_state_field("defaulted")
    distribute = get_property_of_state_field("distribute")
    collect = get_property_of_state_field("collect")
    synch = get_property_of_state_field("synch")
    aggregate = get_property_of_state_field("aggregate")
    timeout = get_property_of_state_field("timeout")
    activity = get_property_of_state_field("activity")


class LACPCollectorInfoTlv(packetracer.Packet):
    __hdr__ = (
        ("type", "B", COLLECTOR_TYPE),
        ("len", "B", LACP_COLLECTOR_LEN),
        ("maxdelay", "H", 10),
        ("reserved", "12s", b"\x00" * 12),
    )


class LACPTerminatorTlv(packetracer.Packet):
    __hdr__ = (
        ("type", "B", TERMINATOR_TYPE),
        ("len", "B", 0),
    )


class LACPReserved(packetracer.Packet):
    __hdr__ = (
        ("reserved", "50s", b"\x00" * 50),
    )


LACP_TLV_CLS = {
    ACTOR_TYPE: LACPActorInfoTlv,
    PARTNER_TYPE: LACPPartnerInfoTlv,
    COLLECTOR_TYPE: LACPCollectorInfoTlv,
    TERMINATOR_TYPE: LACPTerminatorTlv,
}



"""
Secure Sockets Layer / Transport Layer Security.
"""
import logging

from packetracer import packetracer, triggerlist
from packetracer.structcbs import unpack_H, unpack_I

#
# Note from April 2011: cde...@gmail.com added code that parses SSL3/TLS messages
# more in depth.
#
# Jul 2012: afleenor@google.com modified and extended SSL support further.
#

logger = logging.getLogger("packetracer")

# SSLv3/TLS versions
SSL3_V = 0x0300
TLS1_V = 0x0301
TLS11_V = 0x0302
TLS12_V = 0x0303

ssl3_versions_str = {
    SSL3_V: "SSL3",
    TLS1_V: "TLS 1.0",
    TLS11_V: "TLS 1.1",
    TLS12_V: "TLS 1.2"
}

# Alert levels
SSL3_AD_WARNING = 1
SSL3_AD_FATAL = 2
alert_level_str = {
    SSL3_AD_WARNING: "SSL3_AD_WARNING",
    SSL3_AD_FATAL: "SSL3_AD_FATAL"
}

# SSL3 alert descriptions
SSL3_AD_CLOSE_NOTIFY = 0
SSL3_AD_UNEXPECTED_MESSAGE = 10  # fatal
SSL3_AD_BAD_RECORD_MAC = 20  # fatal
SSL3_AD_DECOMPRESSION_FAILURE = 30  # fatal
SSL3_AD_HANDSHAKE_FAILURE = 40  # fatal
SSL3_AD_NO_CERTIFICATE = 41
SSL3_AD_BAD_CERTIFICATE = 42
SSL3_AD_UNSUPPORTED_CERTIFICATE = 43
SSL3_AD_CERTIFICATE_REVOKED = 44
SSL3_AD_CERTIFICATE_EXPIRED = 45
SSL3_AD_CERTIFICATE_UNKNOWN = 46
SSL3_AD_ILLEGAL_PARAMETER = 47  # fatal

# TLS1 alert descriptions
TLS1_AD_DECRYPTION_FAILED = 21
TLS1_AD_RECORD_OVERFLOW = 22
TLS1_AD_UNKNOWN_CA = 48  # fatal
TLS1_AD_ACCESS_DENIED = 49  # fatal
TLS1_AD_DECODE_ERROR = 50  # fatal
TLS1_AD_DECRYPT_ERROR = 51
TLS1_AD_EXPORT_RESTRICTION = 60  # fatal
TLS1_AD_PROTOCOL_VERSION = 70  # fatal
TLS1_AD_INSUFFICIENT_SECURITY = 71  # fatal
TLS1_AD_INTERNAL_ERROR = 80  # fatal
TLS1_AD_USER_CANCELLED = 90
TLS1_AD_NO_RENEGOTIATION = 100
# /* codes 110-114 are from RFC3546 */
TLS1_AD_UNSUPPORTED_EXTENSION = 110
TLS1_AD_CERTIFICATE_UNOBTAINABLE = 111
TLS1_AD_UNRECOGNIZED_NAME = 112
TLS1_AD_BAD_CERTIFICATE_STATUS_RESPONSE = 113
TLS1_AD_BAD_CERTIFICATE_HASH_VALUE = 114
TLS1_AD_UNKNOWN_PSK_IDENTITY = 115  # fatal

# Mapping alert types to strings
alert_description_str = {
    SSL3_AD_CLOSE_NOTIFY: "SSL3_AD_CLOSE_NOTIFY",
    SSL3_AD_UNEXPECTED_MESSAGE: "SSL3_AD_UNEXPECTED_MESSAGE",
    SSL3_AD_BAD_RECORD_MAC: "SSL3_AD_BAD_RECORD_MAC",
    SSL3_AD_DECOMPRESSION_FAILURE: "SSL3_AD_DECOMPRESSION_FAILURE",
    SSL3_AD_HANDSHAKE_FAILURE: "SSL3_AD_HANDSHAKE_FAILURE",
    SSL3_AD_NO_CERTIFICATE: "SSL3_AD_NO_CERTIFICATE",
    SSL3_AD_BAD_CERTIFICATE: "SSL3_AD_BAD_CERTIFICATE",
    SSL3_AD_UNSUPPORTED_CERTIFICATE: "SSL3_AD_UNSUPPORTED_CERTIFICATE",
    SSL3_AD_CERTIFICATE_REVOKED: "SSL3_AD_CERTIFICATE_REVOKED",
    SSL3_AD_CERTIFICATE_EXPIRED: "SSL3_AD_CERTIFICATE_EXPIRED",
    SSL3_AD_CERTIFICATE_UNKNOWN: "SSL3_AD_CERTIFICATE_UNKNOWN",
    SSL3_AD_ILLEGAL_PARAMETER: "SSL3_AD_ILLEGAL_PARAMETER",
    TLS1_AD_DECRYPTION_FAILED: "TLS1_AD_DECRYPTION_FAILED",
    TLS1_AD_RECORD_OVERFLOW: "TLS1_AD_RECORD_OVERFLOW",
    TLS1_AD_UNKNOWN_CA: "TLS1_AD_UNKNOWN_CA",
    TLS1_AD_ACCESS_DENIED: "TLS1_AD_ACCESS_DENIED",
    TLS1_AD_DECODE_ERROR: "TLS1_AD_DECODE_ERROR",
    TLS1_AD_DECRYPT_ERROR: "TLS1_AD_DECRYPT_ERROR",
    TLS1_AD_EXPORT_RESTRICTION: "TLS1_AD_EXPORT_RESTRICTION",
    TLS1_AD_PROTOCOL_VERSION: "TLS1_AD_PROTOCOL_VERSION",
    TLS1_AD_INSUFFICIENT_SECURITY: "TLS1_AD_INSUFFICIENT_SECURITY",
    TLS1_AD_INTERNAL_ERROR: "TLS1_AD_INTERNAL_ERROR",
    TLS1_AD_USER_CANCELLED: "TLS1_AD_USER_CANCELLED",
    TLS1_AD_NO_RENEGOTIATION: "TLS1_AD_NO_RENEGOTIATION",
    TLS1_AD_UNSUPPORTED_EXTENSION: "TLS1_AD_UNSUPPORTED_EXTENSION",
    TLS1_AD_CERTIFICATE_UNOBTAINABLE: "TLS1_AD_CERTIFICATE_UNOBTAINABLE",
    TLS1_AD_UNRECOGNIZED_NAME: "TLS1_AD_UNRECOGNIZED_NAME",
    TLS1_AD_BAD_CERTIFICATE_STATUS_RESPONSE: "TLS1_AD_BAD_CERTIFICATE_STATUS_RESPONSE",
    TLS1_AD_BAD_CERTIFICATE_HASH_VALUE: "TLS1_AD_BAD_CERTIFICATE_HASH_VALUE",
    TLS1_AD_UNKNOWN_PSK_IDENTITY: "TLS1_AD_UNKNOWN_PSK_IDENTITY"
}

# Record types
RECORD_TLS_CHG_CIPHERSPEC = 20
RECORD_TLS_ALERT = 21
RECORD_TLS_HANDSHAKE = 22
RECORD_TLS_HANDSHAKE_HELLO = 22 * 10
RECORD_TLS_HANDSHAKE_DATA = 22 * 100
RECORD_TLS_APPDATA = 23

# Handshake types
HNDS_HELLO_REQ = 0
HNDS_HELLO_CLIENT = 1
HNDS_HELLO_SERVER = 2
HNDS_CERTIFICATE = 11
HNDS_SERVER_KEY_EXCHANGE = 12
HNDS_CERTIFICATE_REQ = 13
HNDS_SERVER_HELLO_DONE = 14
HNDS_CERT_VERIFIY = 15
HNDS_CLIENT_KEY_EXCHANGE = 16
HNDS_FINISHED = 20


class SSL(packetracer.Packet):
    __hdr__ = (
        ("records", None, triggerlist.TriggerList),
    )

    def _dissect(self, buf):
        # logger.debug("parsing SSL")
        # parse all records out of message
        # possible types are Client/Sevrer Hello, Change Cipher Spec etc.
        records = []
        offset = 0
        dlen = len(buf)

        # records is the only header so it's ok to avoid lazy dissecting
        while offset < dlen:
            record_len = unpack_H(buf[offset + 3: offset + 5])[0]
            record = Record(buf[offset: offset + 5 + record_len])
            records.append(record)
            offset += 5 + record_len
        # logger.debug("adding %d records", len(records))

        self.records.extend(records)
        return dlen

    def get_cert_length(self):
        """
        return -- length of certificate content if this is a certificate handshake or 0 if not
        """
        for record in self.records:
            try:
                # logger.debug("%r", record)
                if record.type == RECORD_TLS_HANDSHAKE and record.higher_layer.type == HNDS_CERTIFICATE:
                    # logger.debug("type: %X, sub: %X, len: %r",
                    #	record.type, record.handshake.type, record.handshake.len_i)
                    # Handshake Proto -> Cert -> Length
                    return record.higher_layer.len_i - 3
            except Exception as ex:
                logger.warning(ex)
                pass
        return 0


class Extension(packetracer.Packet):
    """
    Handshake protocol extension
    """
    __hdr__ = (
        ("type", "H", 0),
        ("len", "H", 0)
    )


#
# Record contents
#
class Handshake(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        ("len", "3s", b"\x00" * 3)
    )

    len_i = packetracer.get_property_bytes_num("len", ">I")

    def extract_certificates(self):
        """
        Extracts certificates from a Handshake packet
        Workflow:
            find 1# cert segment(SSL.get_cert_length()) -> collect/assemble until
            cert length collected -> create SSL(tcp_bytes) -> ssl.handshake.extract_certs()
        return -- [cert1, cert2, ...]
        """
        ret = []

        if self.type != HNDS_CERTIFICATE:
            logger.warning("not a certificate handshake: %r", self)
            return ret

        bts_body = self.body_bytes
        certs_len = self.len_i - 3
        # logger.debug("total cert length: %d", certs_len)
        # skip total cert length
        off = 3

        while off < certs_len:
            cert_len = unpack_I(b"\x00" + bts_body[off: off + 3])[0]
            # logger.debug("cert length: %d", cert_len)
            cert_bytes = bts_body[off + 3: off + 3 + cert_len]
            off += 3 + cert_len
            ret.append(cert_bytes)

        return ret


class HandshakeHello(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        # can't use struct here but:
        # int.from_bytes(len, "big")
        ("len", "3s", b"\x00" * 3),
        ("tlsversion", "H", 0x0301),
        ("random", "32s", b"\x00" * 32),
        ("sid_len", "B", 32),
        # variable length
        ("sid", None, b"A" * 32),
        ("ciphersuite", "H", 0x0035),
        ("compression", "B", 0),
        ("ext_len", "H", 0x0000),
        ("extensions", None, triggerlist.TriggerList),
    )

    len_i = packetracer.get_property_bytes_num("len", ">I")

    @staticmethod
    def __parse_extension(buf):
        extensions = []
        offset = 0
        buflen = len(buf)

        while offset < buflen:
            ext_content_len = unpack_H(buf[offset + 2: offset + 4])[0]
            ext_len = 4 + ext_content_len
            extensions.append(Extension(buf[offset: offset + ext_len]))
            offset += ext_len

        return extensions

    def _dissect(self, buf):
        sid_len = buf[38]
        offset_extlen = 38 + sid_len + 3
        # ext_len = unpack_H(buf[offset_extlen : offset_extlen+2])
        self._init_triggerlist("extensions", buf[offset_extlen + 2:], self.__parse_extension)


class HandshakeData(packetracer.Packet):
    pass


class ChangeCipherSpec(packetracer.Packet):
    pass


class ApplicationData(packetracer.Packet):
    pass


class Record(packetracer.Packet):
    """
    SSLv3 or TLSv1+ Record layer.
    """

    __hdr__ = (
        ("type", "B", 0),
        ("version", "H", 0),
        ("len", "H", 0),
    )

    __handler__ = {
        RECORD_TLS_CHG_CIPHERSPEC: ChangeCipherSpec,
        RECORD_TLS_HANDSHAKE: Handshake,
        RECORD_TLS_HANDSHAKE_HELLO: HandshakeHello,
        RECORD_TLS_HANDSHAKE_DATA: HandshakeData,
        RECORD_TLS_APPDATA: ApplicationData
    }

    def _dissect(self, buf):
        # TODO: check for other handshakes
        # if buf[0] == RECORD_TLS_HANDSHAKE:
        #	logger.debug("got a handshake")
        self._init_handler(buf[0], buf[5:])
        return 5



"""
Packet interceptor using NFQueue

Requirements:
- CPython
- NFQUEUE target support in Kernel
- iptables
"""
import ctypes
import errno
import logging
import socket
import threading
from collections import namedtuple
from ctypes import util as utils
from socket import htons, ntohl, ntohs
from socket import timeout as socket_timeout

logger = logging.getLogger("packetracer")

MSG_NO_NFQUEUE = "Could not load netfilter_queue library. See README.md for interceptor requirements."

netfilter = None

try:
    # load library
    nflib = utils.find_library("netfilter_queue")

    if nflib is None:
        raise RuntimeError()

    netfilter = ctypes.cdll.LoadLibrary(nflib)
except:
    logger.warning(MSG_NO_NFQUEUE)


class NfqQHandler(ctypes.Structure):
    pass


class NfnlHandle(ctypes.Structure):
    pass


nfnl_callback_ctype = ctypes.CFUNCTYPE(
    ctypes.c_int, *(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p)
)


class NfnlCallback(ctypes.Structure):
    _fileds_ = [("call", nfnl_callback_ctype),
                ("data", ctypes.c_void_p),
                ("attr_count", ctypes.c_uint16)
                ]


class NfnlSubsysHandle(ctypes.Structure):
    _fields_ = [("nfilter_handler", ctypes.POINTER(NfnlHandle)),
                ("subscriptions", ctypes.c_uint32),
                ("subsys_id", ctypes.c_uint8),
                ("cb_count", ctypes.c_uint8),
                ("callback", ctypes.POINTER(NfnlCallback))
                ]


class NfqHandle(ctypes.Structure):
    _fields_ = [("nfnlh", ctypes.POINTER(NfnlHandle)),
                ("nfnlssh", ctypes.POINTER(NfnlSubsysHandle)),
                ("qh_list", ctypes.POINTER(NfqQHandler))
                ]


class NfqQHandle(ctypes.Structure):
    _fields_ = [("next", ctypes.POINTER(NfqQHandler)),
                ("h", ctypes.POINTER(NfqHandle)),
                ("id", ctypes.c_uint16),
                ("cb", ctypes.POINTER(NfnlHandle)),
                ("data", ctypes.c_void_p)
                ]


class NfqData(ctypes.Structure):
    _fields_ = [("data", ctypes.POINTER(ctypes.c_void_p))]


class NfqnlMsgPacketHw(ctypes.Structure):
    _fields_ = [("hw_addrlen", ctypes.c_uint16),
                ("_pad", ctypes.c_uint16),
                #############################
                ("hw_addr", ctypes.c_uint8 * 8)]


class NfqnlMsgPacketHdr(ctypes.Structure):
    _fields_ = [("packet_id", ctypes.c_uint32),
                ("hw_protocol", ctypes.c_uint16),
                ("hook", ctypes.c_uint8)
                ]


class Timeval(ctypes.Structure):
    _fields_ = [("tv_sec", ctypes.c_long),
                ("tv_usec", ctypes.c_long)]


# Return netfilter netlink handler
nfnlh = netfilter.nfq_nfnlh
nfnlh.restype = ctypes.POINTER(NfnlHandle)
nfnlh.argtypes = ctypes.POINTER(NfqHandle),

# Return a file descriptor for the netlink connection associated with the
# given queue connection handle.
nfq_fd = netfilter.nfnl_fd
nfq_fd.restype = ctypes.c_int
nfq_fd.argtypes = ctypes.POINTER(NfnlHandle),

nfnl_rcvbufsiz = netfilter.nfnl_rcvbufsiz
nfnl_rcvbufsiz.restype = ctypes.c_int
nfnl_rcvbufsiz.argtypes = ctypes.POINTER(NfnlHandle), ctypes.c_uint

# This function obtains a netfilter queue connection handle
ll_open_queue = netfilter.nfq_open
ll_open_queue.restype = ctypes.POINTER(NfqHandle)

# This function closes the nfqueue handler and free associated resources.
close_queue = netfilter.nfq_close
close_queue.restype = ctypes.c_int
close_queue.argtypes = ctypes.POINTER(NfqHandle),

# Bind a nfqueue handler to a given protocol family.
bind_pf = netfilter.nfq_bind_pf
bind_pf.restype = ctypes.c_int
bind_pf.argtypes = ctypes.POINTER(NfqHandle), ctypes.c_uint16

# Unbind nfqueue handler from a protocol family.
unbind_pf = netfilter.nfq_unbind_pf
unbind_pf.restype = ctypes.c_int
unbind_pf.argtypes = ctypes.POINTER(NfqHandle), ctypes.c_uint16

# Creates a new queue handle, and returns it.
create_queue = netfilter.nfq_create_queue
create_queue.restype = ctypes.POINTER(NfqQHandler)
create_queue.argtypes = ctypes.POINTER(NfqHandle), ctypes.c_uint16, ctypes.c_void_p, ctypes.c_void_p

# Removes the binding for the specified queue handle.
destroy_queue = netfilter.nfq_destroy_queue
destroy_queue.restype = ctypes.c_int
destroy_queue.argtypes = ctypes.POINTER(NfqQHandler),

# Triggers an associated callback for the given packet received from the queue.
handle_packet = netfilter.nfq_handle_packet
handle_packet.restype = ctypes.c_int
handle_packet.argtypes = ctypes.POINTER(NfqHandle), ctypes.c_char_p, ctypes.c_int

# nfqnl_config_mode
NFQNL_COPY_NONE, NFQNL_COPY_META, NFQNL_COPY_PACKET = 0, 1, 2

# Sets the amount of data to be copied to userspace for each packet queued
# to the given queue.
#
# NFQNL_COPY_NONE - do not copy any data
# NFQNL_COPY_META - copy only packet metadata
# NFQNL_COPY_PACKET - copy entire packet
set_mode = netfilter.nfq_set_mode
set_mode.restype = ctypes.c_int
set_mode.argtypes = ctypes.POINTER(NfqQHandler), ctypes.c_uint8, ctypes.c_uint

# Sets the size of the queue in kernel. This fixes the maximum number
# of packets the kernel will store before internally before dropping
# upcoming packets.
set_queue_maxlen = netfilter.nfq_set_queue_maxlen
set_queue_maxlen.restype = ctypes.c_int
set_queue_maxlen.argtypes = ctypes.POINTER(NfqQHandler), ctypes.c_uint32

# Responses from hook functions.
NF_DROP, NF_ACCEPT, NF_STOLEN = 0, 1, 2
NF_QUEUE, NF_REPEAT, NF_STOP = 3, 4, 5
NF_MAX_VERDICT = NF_STOP

# Notifies netfilter of the userspace verdict for the given packet. Every
# queued packet _must_ have a verdict specified by userspace, either by
# calling this function, or by calling the nfq_set_verdict_mark() function.
# NF_DROP - Drop packet
# NF_ACCEPT - Accept packet
# NF_STOLEN - Don't continue to process the packet and not deallocate it.
# NF_QUEUE - Enqueue the packet
# NF_REPEAT - Handle the same packet
# NF_STOP -
# NF_MAX_VERDICT -
set_verdict = netfilter.nfq_set_verdict
set_verdict.restype = ctypes.c_int
set_verdict.argtypes = ctypes.POINTER(NfqQHandler), ctypes.c_uint32, ctypes.c_uint32, \
                       ctypes.c_uint32, ctypes.c_char_p

# Return the metaheader that wraps the packet.
get_msg_packet_hdr = netfilter.nfq_get_msg_packet_hdr
get_msg_packet_hdr.restype = ctypes.POINTER(NfqnlMsgPacketHdr)
get_msg_packet_hdr.argtypes = ctypes.POINTER(NfqData),

# Retrieves the hardware address associated with the given queued packet.
# struct nfqnl_msg_packet_hw* nfq_get_packet_hw	(	struct nfq_data * 	nfad	 ) 	[read]
# Can be used to retrieve the source MAC address.
# The destination MAC address is not known until after POSTROUTING and a successful ARP request,
# so cannot currently be retrieved. (nfqueue documentation)
get_packet_hw = netfilter.nfq_get_packet_hw
get_packet_hw.restype = ctypes.POINTER(NfqnlMsgPacketHw)
get_packet_hw.argtypes = ctypes.POINTER(NfqData),

# Retrieve the payload for a queued packet.
get_payload = netfilter.nfq_get_payload
get_payload.restype = ctypes.c_int
get_payload.argtypes = ctypes.POINTER(NfqData), ctypes.POINTER(ctypes.c_void_p)

HANDLER = ctypes.CFUNCTYPE(
    # (struct NfqQHandler *qh, struct nfgenmsg *nfmsg, struct NfqData *nfa, void *data)
    None, *(ctypes.POINTER(NfqQHandler), ctypes.c_void_p, ctypes.POINTER(NfqData), ctypes.c_void_p)
)


def get_full_payload(nfa, ptr_packet):
    len_recv = get_payload(nfa, ctypes.byref(ptr_packet))
    data = ctypes.string_at(ptr_packet, len_recv)
    return len_recv, data


class Interceptor(object):
    """
    Packet interceptor. Allows MITM and filtering.
    Example config for iptables:
    iptables -I INPUT 1 -p icmp -j NFQUEUE --queue-balance 0:2
    """
    QueueConfig = namedtuple("QueueConfig",
                             ["queue", "queue_id", "nfq_handle", "nfq_socket", "verdictthread", "handler"])

    def __init__(self, nfqueue_size=2048, rcvbufsiz=2048):
        """
        nfqueue_size -- Sets the size of the queue in kernel. This fixes the maximum number of packets the
            kernel will store before internally before dropping upcoming packets.
        rcvbufsiz -- Sets the new size of the socket buffer. Use this setting to increase the socket buffer
            size if your system is reporting ENOBUFS errors.
        See: https://www.netfilter.org/projects/libnetfilter_queue/doxygen/

        WARNING: Set nfqueue_size and rcvbufsiz to None (or lower values) if there are any problems on receiving
        """
        self._netfilterqueue_configs = []
        self._is_running = False
        self._nfqueue_size = nfqueue_size
        self._rcvbufsiz = rcvbufsiz

    @staticmethod
    def verdict_trigger_cycler(recv, nfq_handle, obj):
        try:
            while obj._is_running:
                try:
                    # max IP packet size = 65535 bytes
                    bts = recv(65535)
                except socket_timeout:
                    continue
                except OSError as e:
                    # Ignore ENOBUFS errors, we can't handle this anyway
                    # Alternative is to set NETLINK_NO_ENOBUFS socket option
                    if e.errno == errno.ENOBUFS:
                        # logger.debug("Droppin' a packet, consider increasing receive buffer")
                        continue
                    raise e

                handle_packet(nfq_handle, bts, len(bts))
        except OSError:
            # eg "Bad file descriptor": started and nothing read yet
            # logger.error(ex)
            pass
        except Exception as ex:
            logger.error("Exception while reading: %r", ex)

    # finally:
    #	logger.debug("verdict_trigger_cycler finished, stopping Interceptor")
    #	obj.stop()

    def _setup_queue(self, queue_id, ctx, verdict_callback):
        def verdict_callback_ind(queue_handle, nfmsg, nfa, _data):
            packet_ptr = ctypes.c_void_p(0)

            # logger.debug("verdict cb for queue %d", queue_id)
            pkg_hdr = get_msg_packet_hdr(nfa)
            packet_id = ntohl(pkg_hdr.contents.packet_id)
            linklayer_protoid = htons(pkg_hdr.contents.hw_protocol)

            len_recv, data = get_full_payload(nfa, packet_ptr)

            try:
                hw_info = get_packet_hw(nfa).contents
                hw_addrlen = ntohs(hw_info.hw_addrlen)
                hw_addr = ctypes.string_at(hw_info.hw_addr, size=hw_addrlen)
            except:
                # hw address not always present, eg DHCP discover -> offer...
                hw_addr = None

            data_ret, verdict = data, NF_DROP

            try:
                data_ret, verdict = verdict_callback(hw_addr, linklayer_protoid, data, ctx)
            except Exception as ex:
                logger.warning("Verdict callback problem, packet will be dropped: %r", ex)

            set_verdict(queue_handle, packet_id, verdict, len(data_ret), ctypes.c_char_p(data_ret))

        nfq_handle = ll_open_queue()  # 2

        # This call is obsolete, Linux kernels from 3.8 onwards ignore it.
        # unbind_pf(nfq_handle, socket.AF_INET)
        # bind_pf(nfq_handle, socket.AF_INET)

        c_handler = HANDLER(verdict_callback_ind)
        queue = create_queue(nfq_handle, queue_id, c_handler, None)  # 1

        set_mode(queue, NFQNL_COPY_PACKET, 0xFFFF)

        nf = nfnlh(nfq_handle)
        fd = nfq_fd(nf)
        # fd, family, sockettype
        nfq_socket = socket.fromfd(fd, 0, 0)  # 3

        if self._nfqueue_size is not None:
            ret = set_queue_maxlen(queue, self._nfqueue_size)
            if ret == -1:
                logger.warning("Could not set queue_maxlen to %d", self._nfqueue_size)

        if self._rcvbufsiz is not None:
            ret = nfnl_rcvbufsiz(nf, self._rcvbufsiz)
        # logger.debug("Update rcvbufsiz: %d", ret)

        # TODO: better solution to check for running state? close socket and raise exception does not work in stop()
        nfq_socket.settimeout(1)

        thread = threading.Thread(
            target=Interceptor.verdict_trigger_cycler,
            args=[nfq_socket.recv, nfq_handle, self]
        )

        thread.start()

        qconfig = Interceptor.QueueConfig(
            queue=queue, queue_id=queue_id, nfq_handle=nfq_handle, nfq_socket=nfq_socket,
            verdictthread=thread, handler=c_handler
        )
        self._netfilterqueue_configs.append(qconfig)

    def start(self, verdict_callback, queue_ids, ctx=None):
        """
        verdict_callback -- callback with this signature:
            callback(data, ctx): data, verdict
        queue_id -- id of the que placed using iptables
        ctx -- context object given to verdict callback
        """
        if self._is_running:
            return

        if queue_ids is None:
            queue_ids = []

        self._is_running = True

        for queue_id in queue_ids:
            # setup queue and start produces threads
            self._setup_queue(queue_id, ctx, verdict_callback)

    def stop(self):
        if not self._is_running:
            return

        # logger.debug("stopping Interceptor")
        self._is_running = False

        for qconfig in self._netfilterqueue_configs:
            destroy_queue(qconfig.queue)
            close_queue(qconfig.nfq_handle)
            qconfig.nfq_socket.close()
            # logger.debug("joining verdict thread for queue %d", qconfig.queue_id)
            qconfig.verdictthread.join()

        self._netfilterqueue_configs.clear()



"""Ethernet Flow Control"""
from packetracer import packetracer, triggerlist
from packetracer.structcbs import pack_H, unpack_H

PAUSE_OPCODE = 0x0001  # Pause frame IEEE 802.3x
PFC_OPCODE = 0x0101  # Priority Flow Control IEEE 802.1Qbb


class FlowControl(packetracer.Packet):
    __hdr__ = (
        ("opcode", "H", PAUSE_OPCODE),
    )

    def _dissect(self, buf):
        if buf[:2] == b"\x01\x01":
            self._init_handler(PFC_OPCODE, buf[2:])
        else:
            self._init_handler(PAUSE_OPCODE, buf[2:])
        return 2

    class Pause(packetracer.Packet):
        __hdr__ = (
            ("ptime", "H", 0x0000),
        )

    class PFC(packetracer.Packet):
        __hdr__ = (
            ("ms", "B", 0),  # most significant octet is reserved,set to zero
            ("ls", "B", 0),  # least significant octet indicates time_vector parameter
            ("time", None, triggerlist.TriggerList),
        )

        # Conveniant access to ls field(bit representation via list)
        # e.g. 221 -> [1, 1, 0, 1, 1, 1, 0, 1]
        def __get_ls(self):
            # return [(self.ls >> x) & 1 for x in reversed(range(8))]
            return [int(bstr) for bstr in bin(self.ls)[2:]]

        # e.g. [1, 1, 0, 1, 1, 1, 0, 1] -> 221
        def __set_ls(self, value):
            # self.ls = int("".join(map(str, value)), 2)
            self.ls = int("".join(["%d" % bint for bint in value]), 2)

        ls_list = property(__get_ls, __set_ls)

        # Conveniant access to time field(decimal representation via list)
        def __get_time(self):
            return [unpack_H(x)[0] for x in self.time]

        def __set_time(self, value):
            self.time = [pack_H(x) for x in value]

        time_list = property(__get_time, __set_time)

        def _dissect(self, buf):
            for i in range(2, 18, 2):
                self.time.append(buf[i:i + 2])
            # TODO: find more efficient way, always correct?
            return 2 + len(self.time) * 2

    __handler__ = {
        PAUSE_OPCODE: Pause,
        PFC_OPCODE: PFC
    }



"""Internetwork Packet Exchange."""

from packetracer import packetracer

IPX_HDR_LEN = 30


class IPX(packetracer.Packet):
    __hdr__ = (
        ("sum", "H", 0xFFFF),
        ("len", "H", IPX_HDR_LEN),
        ("tc", "B", 0),
        ("pt", "B", 0),
        ("dst", "12s", b""),
        ("src", "12s", b"")
    )



"""
User Datagram Protocol (UDP)

RFC 768 - User Datagram Protocol
RFC 2460 - Internet Protocol, Version 6 (IPv6) Specification
RFC 2675 - IPv6 Jumbograms
RFC 4113 - Management Information Base for the UDP
RFC 5405 - Unicast UDP Usage Guidelines for Application Designers
"""
import logging
import struct

from packetracer import packetracer, checksum
# handler
from packetracer.layer567 import telnet, tftp, dns, dhcp, iso15118, ntp, rtp, sip, pmap, radius, stun
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE, FIELD_FLAG_IS_TYPEFIELD
from packetracer.structcbs import unpack_H, pack_ipv4_header, pack_ipv6_header

# avoid references for performance reasons
in_cksum = checksum.in_cksum

logger = logging.getLogger("packetracer")

UDP_PORT_MAX = 65535

UDP_PROTO_TELNET = 23
UDP_PROTO_DNS = (53, 5353)
UDP_PROTO_DHCP = (67, 68)
UDP_PROTO_TFTP = 69
UDP_PROTO_PMAP = 111
UDP_PROTO_NTP = 123
UDP_PROTO_RADIUS = (1812, 1813, 1645, 1646)
UDP_PROTO_STUN = 3478
UDP_PROTO_RTP = (5004, 5005)
UDP_PROTO_SIP = (5060, 5061)
UDP_PROTO_ISO15118 = 15118


class UDP(packetracer.Packet):
    __hdr__ = (
        ("sport", "H", 0xDEAD),
        ("dport", "H", 0, FIELD_FLAG_AUTOUPDATE | FIELD_FLAG_IS_TYPEFIELD),
        ("ulen", "H", 8, FIELD_FLAG_AUTOUPDATE),  # header + body, min 8
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE)
    )

    __handler__ = {
        UDP_PROTO_TELNET: telnet.Telnet,
        UDP_PROTO_TFTP: tftp.TFTP,
        UDP_PROTO_DNS: dns.DNS,
        UDP_PROTO_DHCP: dhcp.DHCP,
        UDP_PROTO_ISO15118: iso15118.SDP,
        UDP_PROTO_PMAP: pmap.Pmap,
        UDP_PROTO_NTP: ntp.NTP,
        UDP_PROTO_RADIUS: radius.Radius,
        UDP_PROTO_RTP: rtp.RTP,
        UDP_PROTO_SIP: sip.SIP,
        UDP_PROTO_STUN: stun.STUN
    }

    def _update_fields(self):
        # UDP-checksum needs to be updated on one of the following:
        # - this layer itself or any upper layer changed
        # - changes to the IP-pseudoheader
        # There is no update on user-set checksums.
        # changed = self._changed()
        update = True

        if self.ulen_au_active:
            self.ulen = len(self)

        # self._update_higherlayer_id()

        try:
            # changes to IP-layer, don't mind if this isn't IP
            if not self._lower_layer._header_changed:
                # lower layer doesn't need update, check for changes in present and upper layer
                # logger.debug("lower layer did NOT change!")
                update = True
        except AttributeError:
            # assume not an IP packet: we can't calculate the checksum
            update = False

        if update and self.sum_au_active:
            self._calc_sum()

    def _dissect(self, buf):
        ports = [unpack_H(buf[0:2])[0], unpack_H(buf[2:4])[0]]

        try:
            # source or destination port should match
            htype = [x for x in ports if x in packetracer.Packet._id_handlerclass_dct[UDP]][0]
            self._init_handler(htype, buf[8:])
        except:
            # no type found
            # logger.debug("could not parse type: %d because: %s" % (type, e))
            pass
        return 8

    def _calc_sum(self):
        """Recalculate the UDP-checksum."""
        # TCP and underwriting are freaky bitches: we need the IP pseudoheader to calculate their checksum
        # logger.debug("UDP sum recalc, sport=%s/dport=%s" % (self.sport, self.dport))
        try:
            # we need src/dst for checksum-calculation
            src, dst = self._lower_layer.src, self._lower_layer.dst
            # logger.debug(src + b" / "+ dst)
            self.sum = 0
            udp_bin = self.header_bytes + self.body_bytes

            # IP-pseudoheader, check if version 4 or 6
            if len(src) == 4:
                s = pack_ipv4_header(src, dst, 17, len(udp_bin))  # 17 = UDP
            else:
                s = pack_ipv6_header(src, dst, 17, len(udp_bin))  # 17 = UDP

            csum = in_cksum(s + udp_bin)

            if csum == 0:
                csum = 0xFFFF  # RFC 768, p2

            # get the checksum of concatenated pseudoheader+TCP packet
            # assign via non-shadowed variable to trigger re-packing
            self.sum = csum
        except (AttributeError, struct.error):
            # not an IP packet as lower layer (src, dst not present) or invalid src/dst
            pass

    def direction(self, other):
        direction = 0
        # logger.debug("checking direction: %s<->%s" % (self, other))
        if self.sport == other.sport and self.dport == other.dport:
            direction = packetracer.Packet.DIR_SAME
        if self.sport == other.dport and self.dport == other.sport:
            direction = packetracer.Packet.DIR_REV
        if direction == 0:
            return packetracer.Packet.DIR_UNKNOWN
        return direction

    def reverse_address(self):
        self.sport, self.dport = self.dport, self.sport



"""
Simple socket wrapper for reading/writing on layer 2.
For all other use cases standard python sockets
should be used.
"""
import logging
import socket
import ssl

from packetracer import packetracer
from packetracer.layer12 import ethernet

logger = logging.getLogger("packetracer")


class SocketHndl(object):
    """
    Simple socket handler for layer 2 reading/writing.
    """
    ETH_P_ALL = 0x0003
    ETH_P_IPV4 = 0x0800

    def __init__(self, iface_name="lo",
                 timeout=3,
                 buffersize_recv=None,
                 buffersize_send=None, **params):
        """
        iface_name -- Bind to the given interface
        timeout -- read timeout in seconds
        buffersize_recv, buffersize_send -- amount of bytes used for receiving and sending
        """

        self.iface_name = iface_name
        self._socket = None
        # man 7 raw -> Receiving of all IP protocols via IPPROTO_RAW
        # is not possible using raw sockets.
        # socket(AF_INET, SOCK_RAW, IPPROTO_RAW)

        logger.info("creating socket, interface to bind on: %s", iface_name)
        try:
            self._socket = socket.socket(socket.AF_PACKET,
                                         socket.SOCK_RAW,
                                         socket.htons(SocketHndl.ETH_P_ALL))
        except OSError as err:
            logger.warning(err)
            logger.warning("Reducing receive scope to IPv4-only")
            self._socket = socket.socket(socket.AF_PACKET,
                                         socket.SOCK_RAW,
                                         socket.htons(SocketHndl.ETH_P_IPV4))

        if iface_name is not None:
            self._socket.bind((iface_name, SocketHndl.ETH_P_ALL))

        self._socket.settimeout(timeout)

        if buffersize_recv is not None:
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, buffersize_recv)
        if buffersize_send is not None:
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, buffersize_send)

    def send(self, bts):
        """
        Send the given bytes to network.

        bts -- the bytes to be sent
        """
        self._socket.send(bts)

    def recv(self, size=65536):
        """
        return -- bytes received from network
        """
        return self._socket.recv(size)

    def __enter__(self):
        return self

    def __exit__(self, objtype, value, traceback):
        self.close()

    def __iter__(self):
        """
        Call __next__() until StopIteration
        """
        try:
            while True:
                yield self.__next__()
        except StopIteration:
            return

    def __next__(self):
        try:
            return self.recv()
        except socket.timeout:
            raise StopIteration

    def recvp(self, filter_match_recv=lambda _: True, lowest_layer=ethernet.Ethernet, max_amount=1):
        """
        Receive packets from network. This does the same as calling recv() but using a receive
        filter and received bytes will be converted to packets using class given by lowest_layer.
        Raises socket.timeout on timeout

        filter_match_recv -- filter as callback function to match packets to be retrieved.
            Callback-structure: fct(packet), Return True to accept a specific packet.
            Raise StopIteration to stop receiving packets, max_amount will match after all.
        lowest_layer -- packet class to be used to create new packets
        max_amount -- maximum amount of packets to be fetched
        return -- packets received from network as list
        """
        received = []
        # logger.debug("listening for packets")

        while len(received) < max_amount:
            bts = self.recv()
            packet_recv = lowest_layer(bts)
            # logger.debug("got packet: %s" % packet_recv)
            try:
                if filter_match_recv(packet_recv):
                    received.append(packet_recv)
            except StopIteration:
                break
            except:
                # any other exception: ignore
                pass

        return received

    def recvp_iter(self, filter_match_recv=lambda _: True, lowest_layer=ethernet.Ethernet):
        """
        Same as recvp but using iterator returning one packet per cycle.
        """
        while True:
            try:
                bts = self.recv()
            except socket.timeout:
                return

            packet_recv = lowest_layer(bts)
            # logger.debug("got packet: %s" % packet_recv)
            try:
                if filter_match_recv(packet_recv):
                    yield packet_recv
            except StopIteration:
                return
            except:
                continue

    def sr(self, packet_send, max_packets_recv=1, pfilter=lambda _: True, lowest_layer=ethernet.Ethernet):
        """
        Send a packet and receive answer packets. This will use information retrieved
        from direction() to retrieve answer packets. This is not 100% reliable as
        it primarily depends on source/destination data of layers like Ethernet, IP etc.
        Raises socket.timeout on timeout.

        packet_send -- packetracer packet to be sent
        max_packets_recv -- max packets to be received
        pfilter -- filter as lambda function to match packets to be retrieved,
            return True to accept a specific packet.
        lowest_layer -- packet class to be used to create new packets

        return -- packets receives
        """

        received = []
        packet_send_clz = packet_send.__class__

        self.send(packet_send.bin())

        while len(received) < max_packets_recv:
            bts = self.recv()
            packet_recv = lowest_layer(bts)
            # logger.debug("got packet: %s" % packet_recv)
            if not pfilter(packet_recv):
                # filter didn't match
                continue

            # start to compare on corresponding receive-layer
            if packet_send.is_direction(packet_recv[packet_send_clz], packetracer.Packet.DIR_REV):
                # logger.debug("direction matched: %s" % packet_recv)
                received.append(packet_recv)

        return received

    def close(self):
        try:
            self._socket.close()
        except:
            pass


def get_ssl_clientsocket(hostname, port, timeout=5, **sslparams):
    """
    return -- SSL wrapped TCP socket, not complaining about any server certificates
    """
    context = ssl.create_default_context()
    socket_simple = socket.create_connection((hostname, port))
    socket_ssl = context.wrap_socket(socket_simple, server_hostname=hostname, **sslparams)
    socket_ssl.settimeout(timeout)
    return socket_ssl


def get_ssl_serversocket(file_certchain, file_privatekey, bindoptions, password_privkey=None):
    """
    Create a SSL based server socket. Useage:
    conn, addr = ssock.accept()
    data = conn.recv()
    conn.send(data)

    Certificate/private key can be create via:
    openssl req -x509 -newkey rsa:4096 -keyout key.pem -out cert.pem -days 365

    return -- SSL wrapped TCP server socket
    """
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(file_certchain, file_privatekey, password=password_privkey)

    socket_simple = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    socket_simple.bind(bindoptions)
    socket_simple.listen(5)

    return context.wrap_socket(socket_simple, server_side=True)


# TCP/UDP server and client example code
# Server
"""
# ncat 127.0.0.1 80
serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # IPv6: AF_INET6
serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
serversocket.bind(("127.0.0.1", 80))
serversocket.listen(5)
(ssock, address) = serversocket.accept()
data = ssock.recv(1024)
print(data)
ssock.send(data)
"""

"""
# ncat 127.0.0.1 80 -u
serversocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # IPv6: AF_INET6
serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
serversocket.bind(("127.0.0.1", 80))
data, address = serversocket.recvfrom(1024)
print(data)
serversocket.sendto(data, address)
"""

# Client
"""
# ncat -l 80
s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # IPv6: AF_INET6
s.connect(("127.0.0.1", 80))
data = s.recv(1024)
print(data)
s.send(data)
"""

"""
# ncat -l 80 -u
s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # IPv6: AF_INET6
s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
#s.bind(("127.0.0.1", 1234))
s.sendto(b"testtest", ("127.0.0.1", 80))
data, address = s.recvfrom(1024)
print(data)
"""



from struct import Struct

unpack_B = Struct(">B").unpack
pack_B = Struct(">B").pack
unpack_BB = Struct(">BB").unpack

unpack_H = Struct(">H").unpack
pack_H = Struct(">H").pack
unpack_H_le = Struct("<H").unpack
pack_H_le = Struct("<H").pack
unpack_HH = Struct(">HH").unpack
unpack_HHHH = Struct(">HHHH").unpack

unpack_I = Struct(">I").unpack
pack_I = Struct(">I").pack
unpack_I_le = Struct("<I").unpack
pack_I_le = Struct("<I").pack
unpack_I_n = Struct("=I").unpack
unpack_IIII = Struct(">IIII").unpack
pack_IIII = Struct(">IIII").pack
unpack_IIII_le = Struct("<IIII").unpack

pack_L_le = Struct("<L")

unpack_Q_le = Struct("<Q").unpack
unpack_Q = Struct(">Q").unpack
pack_Q = Struct(">Q").pack

pack_ipv4_header = Struct(">4s4sxBH").pack
pack_ipv6_header = Struct(">16s16sxBH").pack

pack_ipv4 = Struct("BBBB").pack
unpack_ipv4 = Struct("BBBB").unpack
pack_mac = Struct("BBBBBB").pack
unpack_mac = Struct("BBBBBB").unpack



"""IEEE 802.11"""
import logging

from packetracer import packetracer
from packetracer import triggerlist
from packetracer import utils
from packetracer.structcbs import unpack_H, pack_Q, unpack_Q_le

logger = logging.getLogger("packetracer")

# Frame Types
MGMT_TYPE = 0
CTL_TYPE = 1
DATA_TYPE = 2

# Frame Sub-Types
# MGMT_TYPE
M_ASSOC_REQ = 0
M_ASSOC_RESP = 1
M_REASSOC_REQ = 2
M_REASSOC_RESP = 3
M_PROBE_REQ = 4
M_PROBE_RESP = 5
M_DISASSOC = 10
M_AUTH = 11
M_DEAUTH = 12
M_ACTION = 13
M_BEACON = 8
M_ATIM = 9

# CTL_TYPE
C_BLOCK_ACK_REQ = 8
C_BLOCK_ACK = 9
C_PS_POLL = 10
C_RTS = 11
C_CTS = 12
C_ACK = 13
C_CF_END = 14
C_CF_END_ACK = 15

# DATA_TYPE
D_NORMAL = 0
D_DATA_CF_ACK = 1
D_DATA_CF_POLL = 2
D_DATA_CF_ACK_POLL = 3
D_NULL = 4
D_CF_ACK = 5
D_CF_POLL = 6
D_CF_ACK_POLL = 7
D_QOS_DATA = 8
D_QOS_CF_ACK = 9
D_QOS_CF_POLL = 10
D_QOS_CF_ACK_POLL = 11
D_QOS_NULL = 12
D_QOS_CF_POLL_EMPTY = 14

TO_DS_FLAG = 1
FROM_DS_FLAG = 2
INTER_DS_FLAG = 3

# name : (mask, offset)
_FRAMECTRL_SUBHEADERDATA = {
    "version": (0x0300, 8),
    "type": (0x0C00, 10),
    "subtype": (0xF000, 12),
    "to_ds": (0x0001, 0),
    "from_ds": (0x0002, 1),
    "more_frag": (0x0004, 2),
    "retry": (0x0008, 3),
    "pwr_mgt": (0x0010, 4),
    "more_data": (0x0020, 5),
    "protected": (0x0040, 6),
    "order": (0x0080, 7),
    "from_to_ds": (0x0002 | 0x0001, 0),
}

# needed to distinguish subtypes via types
TYPE_FACTORS = [16, 32, 64]
TYPE_FACTOR_PROTECTED = 128

_subheader_properties = []

IEEE_FIELDS_SRC_DST_BSSID = ["src", "dst", "bssid"]

# set properties to access flags
for subfield_name, mask_off in _FRAMECTRL_SUBHEADERDATA.items():
    # logger.debug("setting prop: %r, %X, %X" % (subfield_name, mask_off[0], mask_off[1]))
    subheader = [
        subfield_name,
        # lambda**2: avoid lexical closure, do not refer to value via reference
        (lambda mask, off:
         (lambda _obj: (_obj.framectl & mask) >> off))(mask_off[0], mask_off[1]),
        (lambda mask, off:
         (lambda _obj, _val: _obj.__setattr__("framectl",
                                              (_obj.framectl & ~mask) | (_val << off))))(mask_off[0], mask_off[1]),
    ]
    _subheader_properties.append(subheader)


class IEEE80211(packetracer.Packet):
    __hdr__ = (
        # AAAABBCC | 00000000
        # AAAA = subtype BB = type CC = version
        ("framectl", "H", 0),
        ("duration", "H", 0x3A01)  # 314 microseconds
    )

    __hdr_sub__ = _subheader_properties

    def _dissect(self, buf):
        # logger.debug("Starting to dissect")
        # self.type/self.subtype use self.framectl, no unpack will happen in dissect so this has
        # to be done manually
        self.framectl = unpack_H(buf[0:2])[0]
        # logger.debug("ieee80211 bytes=%X, type/subtype is=%X/%X, handler=%r" %
        #	(self.framectl, self.type, self.subtype,
        #	packetracer.Packet._id_handlerclass_dct[self.__class__][TYPE_FACTORS[self.type] + self.subtype]))
        self._init_handler(TYPE_FACTORS[self.type] + self.subtype, buf[4:])
        # logger.debug("Finished IEEE dissect")
        return 4

    def is_beacon(self):
        """return -- True if packet is a beacon. Avoids parsing upper layer."""
        return self.type == MGMT_TYPE and self.subtype == M_BEACON

    def extract_client_macs(self):
        """
        Extracts client MACs from upper layer if this is a data packet.

        return -- [mac_client1, ...] or [] if no client macs could be found
        """
        macs_clients = []

        # data: client -> AP or client <- AP
        if self.type == DATA_TYPE:
            if self.from_ds == 1 and self.to_ds == 0:
                macs_clients.append(self.higher_layer.dst)
            elif self.from_ds == 0 and self.to_ds == 1:
                macs_clients.append(self.higher_layer.src)

        return [addr for addr in macs_clients if not utils.is_special_mac(addr)]

    #
    # mgmt frames
    #
    class Beacon(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            # 12 Bits: 0->4095 | 4 Bits
            # SF SS (LE)
            ("seq_frag", "H", 0),
            # _ts (integer) is saved as LE
            ("_ts", "Q", 0),
            ("interval", "H", 0x6400),
            ("capa", "H", 0x0100),
            ("params", None, triggerlist.TriggerList)
        )

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        def _get_ts(self):
            # LE->BE: dirty but simple
            return unpack_Q_le(pack_Q(self._ts))[0]

        def _set_ts(self, val):
            self._ts = unpack_Q_le(pack_Q(val))[0]

        seq = property(_get_seq, _set_seq)
        ts = property(_get_ts, _set_ts)
        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _dissect(self, buf):
            self._init_triggerlist("params", buf[32:], IEEE80211._unpack_ies)
            return len(buf)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class Action(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("category", "B", 0),
            ("code", "B", 0)
        )

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        class BlockAckRequest(packetracer.Packet):
            __hdr__ = (
                ("dialog", "B", 0),
                ("parameters", "H", 0),
                ("timeout", "H", 0),
                ("starting_seq", "H", 0),
            )

        class BlockAckResponse(packetracer.Packet):
            __hdr__ = (
                ("dialog", "B", 0),
                ("status_code", "H", 0),
                ("parameters", "H", 0),
                ("timeout", "H", 0),
            )

        CATEGORY_BLOCK_ACK = 3
        CODE_BLOCK_ACK_REQUEST = 0
        CODE_BLOCK_ACK_RESPONSE = 1

        dst_s = packetracer.get_property_mac("dst")
        src_s = packetracer.get_property_mac("src")
        bssid_s = packetracer.get_property_mac("bssid")

        def _dissect(self, buf):
            # logger.debug(">>>>>>>> ACTION!!!")
            # category: block ack, code: request or response
            self._init_handler(buf[20] * 4 + buf[21], buf[22:])
            return 22

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class ProbeReq(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("params", None, triggerlist.TriggerList)
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def _dissect(self, buf):
            self._init_triggerlist("params", buf[20:], IEEE80211._unpack_ies)
            return len(buf)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class ProbeResp(Beacon):
        pass

    class AssocReq(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("capa", "H", 0),
            ("interval", "H", 0),
            ("params", None, triggerlist.TriggerList)
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def _dissect(self, buf):
            self._init_triggerlist("params", buf[24:], IEEE80211._unpack_ies)
            return len(buf)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class AssocResp(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("capa", "H", 0),
            ("status", "H", 0),
            ("aid", "H", 0),
            ("params", None, triggerlist.TriggerList)
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def _dissect(self, buf):
            self._init_triggerlist("params", buf[26:], IEEE80211._unpack_ies)
            return len(buf)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class Disassoc(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("reason", "H", 0),
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class ReassocReq(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("capa", "H", 0),
            ("interval", "H", 0),
            ("current_ap", "6s", b"\x00" * 6)
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class Auth(packetracer.Packet):
        """Authentication request."""
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("algo", "H", 0),
            ("authseq", "H", 0x0100),
            ("status", "H", 0)
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class Deauth(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\xff" * 6),
            ("src", "6s", b"\x00" * 6),
            ("bssid", "6s", b"\xff" * 6),
            ("seq_frag", "H", 0),
            ("reason", "H", 0x0700)  # class 3 frame received from non associated client
        )

        dst_s = packetracer.get_property_mac("dst")
        bssid_s = packetracer.get_property_mac("bssid")
        src_s = packetracer.get_property_mac("src")

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    m_decoder = {
        M_BEACON: Beacon,
        M_ACTION: Action,
        M_ASSOC_REQ: AssocReq,
        M_ASSOC_RESP: AssocResp,
        M_DISASSOC: Disassoc,
        M_REASSOC_REQ: ReassocReq,
        M_REASSOC_RESP: AssocResp,
        M_AUTH: Auth,
        M_PROBE_REQ: ProbeReq,
        M_PROBE_RESP: ProbeResp,
        M_DEAUTH: Deauth
    }

    #
    # Control frames: no need for extra layer: 802.11 Base data is enough
    #

    class RTS(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6)
        )

        dst_s = packetracer.get_property_mac("dst")
        src_s = packetracer.get_property_mac("src")

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class CTS(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
        )

        dst_s = packetracer.get_property_mac("dst")

    class ACK(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
        )

        dst_s = packetracer.get_property_mac("dst")

    class BlockAckReq(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("reqctrl", "H", 0),
            ("seq", "H", 0)
        )

        dst_s = packetracer.get_property_mac("dst")
        src_s = packetracer.get_property_mac("src")

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class BlockAck(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
            ("reqctrl", "H", 0),
            ("seq", "H", 0),
            ("bitmap", "Q", 0)
        )

        dst_s = packetracer.get_property_mac("dst")
        src_s = packetracer.get_property_mac("src")

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    class CFEnd(packetracer.Packet):
        __hdr__ = (
            ("dst", "6s", b"\x00" * 6),
            ("src", "6s", b"\x00" * 6),
        )

        dst_s = packetracer.get_property_mac("dst")
        src_s = packetracer.get_property_mac("src")

        def reverse_address(self):
            self.dst, self.src = self.src, self.dst

    c_decoder = {
        C_RTS: RTS,
        C_CTS: CTS,
        C_ACK: ACK,
        C_BLOCK_ACK_REQ: BlockAckReq,
        C_BLOCK_ACK: BlockAck,
        C_CF_END: CFEnd
    }

    #
    # Data frames
    #
    class Dataframe(packetracer.Packet):
        """
        DataFrames need special care: there are too many types of field combinations
        to create classes for every one. Solution: initiate giving lower type "subType"
        via constructor. In order to use "src/dst/bssid" instead of addrX set from_to_ds
        of "subType" to one of the following values:

        [Bit 0: from DS][Bit 1: to DS] = [order of fields]

        00 = 0 = dst, src, bssid
        01 = 1 = bssid, src, dst
        10 = 2 = dst, bssid, src
        11 = 3 = RA, TA, DA, SA
        """

        def __init__(self, *arg, **kwargs):
            if len(arg) > 1:
                # logger.debug("extracting lower layer type: %r" % arg[1])
                self.dtype = arg[1]
            else:
                self.dtype = self
                self._from_to_ds_value = 0
            # logger.debug("dstype: %r" % self.dtype.from_to_ds)
            super().__init__(*arg, **kwargs)

        __hdr__ = (
            ("addr1", "6s", b"\x00" * 6),
            ("addr2", "6s", b"\x00" * 6),
            ("addr3", "6s", b"\x00" * 6),
            ("seq_frag", "H", 0),
            ("addr4", "6s", None),  # to/from-DS = 1
            ("qos_ctrl", "H", 0),  # QoS
            ("sec_param", "Q", 0)  # protected
        )

        def _get_seq(self):
            return (self.seq_frag & 0xFF) << 4 | (self.seq_frag >> 12)

        def _set_seq(self, val):
            self.seq_frag = (val & 0xF) << 12 | (val & 0xFF0) >> 4 | (self.seq_frag & 0x0F00)

        seq = property(_get_seq, _set_seq)

        def reverse_address(self):
            if self.dtype.from_to_ds == 0:
                self.addr1, self.addr2 = self.addr2, self.addr1
            elif self.dtype.from_to_ds == 1:
                self.addr2, self.addr3 = self.addr3, self.addr2
            elif self.dtype.from_to_ds == 2:
                self.addr1, self.addr3 = self.addr3, self.addr1

        def _get_from_to_ds(self):
            return self._from_to_ds_value

        # same property structure as in IEEE80211 class
        from_to_ds = property(_get_from_to_ds)

        # FromDs, ToDS
        # 00 = dst, src, bssid
        # 01 = bssid, src, dst
        # 10 = dst, bssid, src
        # 11 = RA, TA, DA, SA

        def __get_src(self):
            return self.addr2 if self.dtype.from_to_ds in [0, 1] else self.addr3

        def __set_src(self, src):
            if self.dtype.from_to_ds in [0, 1]:
                self.addr2 = src
            else:
                self.addr3 = src

        def __get_dst(self):
            return self.addr1 if self.dtype.from_to_ds in [0, 2] else self.addr3

        def __set_dst(self, dst):
            if self.dtype.from_to_ds in [0, 2]:
                self.addr1 = dst
            else:
                self.addr3 = dst

        def __get_bssid(self):
            dstype = self.dtype.from_to_ds

            if dstype == 0:
                return self.addr3
            elif dstype == 1:
                return self.addr1
            elif dstype == 2:
                return self.addr2

        def __set_bssid(self, bssid):
            dstype = self.dtype.from_to_ds
            if dstype == 0:
                self.addr3 = bssid
            elif dstype == 1:
                self.addr1 = bssid
            elif dstype == 2:
                self.addr2 = bssid

        src = property(__get_src, __set_src)
        src_s = packetracer.get_property_mac("src")
        dst = property(__get_dst, __set_dst)
        dst_s = packetracer.get_property_mac("dst")
        bssid = property(__get_bssid, __set_bssid)
        bssid_s = packetracer.get_property_mac("bssid")

        __QOS_SUBTYPES = {8, 9, 10, 11, 12, 14, 15}

        def _dissect(self, buf):
            # logger.debug("starting dissecting, buflen: %r" % str(buf))
            header_len = 30

            try:
                is_qos = True if self.dtype.subtype in IEEE80211.Dataframe.__QOS_SUBTYPES else False
                is_protected = self.dtype.protected == 1
                is_bridge = True if self.dtype.from_ds == 1 and self.dtype.to_ds == 1 else False
            except Exception:
                # logger.debug(e)
                # default is fromds
                is_qos = False
                is_protected = False
                is_bridge = False

            # logger.debug("switching fields1")
            if not is_qos:
                self.qos_ctrl = None
                header_len -= 2
            # logger.debug("switching fields2")
            if not is_protected:
                self.sec_param = None
                header_len -= 8
            # logger.debug("switching fields3")
            if is_bridge:
                self.addr4 = b"\x00" * 6
                header_len += 6
            # logger.debug("format/length/len(bin): %s/%d/%d" % (self._hdr_fmtstr, self.hdr_len, len(self.bin())))
            # logger.debug("%r" % self)
            return header_len

    d_decoder = {
        D_NORMAL: Dataframe,
        D_DATA_CF_ACK: Dataframe,
        D_DATA_CF_POLL: Dataframe,
        D_DATA_CF_ACK_POLL: Dataframe,
        D_NULL: Dataframe,
        D_CF_ACK: Dataframe,
        D_CF_POLL: Dataframe,
        D_CF_ACK_POLL: Dataframe,
        D_QOS_DATA: Dataframe,
        D_QOS_CF_ACK: Dataframe,
        D_QOS_CF_POLL: Dataframe,
        D_QOS_CF_ACK_POLL: Dataframe,
        D_QOS_NULL: Dataframe,
        D_QOS_CF_POLL_EMPTY: Dataframe
    }

    #
    # IEs for Mgmt-Frames
    #
    @staticmethod
    def _unpack_ies(buf):
        """Parse IEs and return them as Triggerlist."""
        # each IE starts with an ID and a length
        ies = []
        off = 0
        buflen = len(buf)
        # logger.debug("lazy dissecting: %s" % buf)

        while off + 2 < buflen:
            ie_id = buf[off]
            try:
                parser = IEEE80211.ie_decoder[ie_id]
            except KeyError:
                # some unknown tag, use standard format
                parser = IEEE80211.IE

            dlen = buf[off + 1]
            # logger.debug("IE parser is: %d = %s = %s" % (ie_id, parser, buf[off: off+2+dlen]))
            try:
                ie = parser(buf[off: off + 2 + dlen])
                ies.append(ie)
            except:
                # Not enough bytes for handler, add raw bytes
                ies.append(buf[off: off + 2 + dlen])
            off += 2 + dlen
        # logger.debug("Finished IE parsing")

        return ies

    class IE(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0)
        )

    class FH(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0),
            ("tu", "H", 0),
            ("hopset", "B", 0),
            ("hoppattern", "B", 0),
            ("hopindex", "B", 0)
        )

    class DS(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0),
            ("ch", "B", 0)
        )

    class CF(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0),
            ("count", "B", 0),
            ("period", "B", 0),
            ("max", "H", 0),
            ("dur", "H", 0)
        )

    class TIM(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0),
            ("count", "B", 0),
            ("period", "B", 0),
            ("ctrl", "H", 0)
        )

    class IBSS(packetracer.Packet):
        __hdr__ = (
            ("id", "B", 0),
            ("len", "B", 0),
            ("atim", "H", 0)
        )

    # IEs
    IE_SSID = 0
    IE_RATES = 1
    IE_FH = 2
    IE_DS = 3
    IE_CF = 4
    IE_TIM = 5
    IE_IBSS = 6
    IE_HT_CAPA = 45
    IE_ESR = 50
    IE_HT_INFO = 61

    ie_decoder = {
        IE_SSID: IE,
        IE_RATES: IE,
        IE_FH: FH,
        IE_DS: DS,
        IE_CF: CF,
        IE_TIM: TIM,
        IE_IBSS: IBSS,
        IE_HT_CAPA: IE,
        IE_ESR: IE,
        IE_HT_INFO: IE
    }


# handler for IEEE80211
# position in list = type-ID
dicts = [IEEE80211.m_decoder, IEEE80211.c_decoder, IEEE80211.d_decoder]
decoder_dict_complete = {}

for pos, decoder_dict in enumerate(dicts):
    for key_decoder, val_decoder in decoder_dict.items():
        # same subtype-ID for different type-IDs, distinguish via "type_factor + subtype"
        # not doing so would lead to eg: type:0 + subtype:1 == type:1 + subtype:0
        decoder_dict_complete[TYPE_FACTORS[pos] + key_decoder] = val_decoder

packetracer.Packet.load_handler(IEEE80211, decoder_dict_complete)

# handler for Action
CATEGORY_BLOCK_ACK_FACTOR = IEEE80211.Action.CATEGORY_BLOCK_ACK * 4
packetracer.Packet.load_handler(IEEE80211.Action,
                                {
                                    CATEGORY_BLOCK_ACK_FACTOR + IEEE80211.Action.CODE_BLOCK_ACK_REQUEST: IEEE80211.Action.BlockAckRequest,
                                    CATEGORY_BLOCK_ACK_FACTOR + IEEE80211.Action.CODE_BLOCK_ACK_RESPONSE: IEEE80211.Action.BlockAckResponse
                                }
                                )



"""Routing Information Protocol."""
import logging

from packetracer import packetracer, triggerlist

logger = logging.getLogger("packetracer")

# RIP v2 - RFC 2453
# http://tools.ietf.org/html/rfc2453

REQUEST = 1
RESPONSE = 2


class RIP(packetracer.Packet):
    __hdr__ = (
        ("cmd", "B", REQUEST),
        ("v", "B", 2),
        ("rsvd", "H", 0),
        ("rte_auth", None, triggerlist.TriggerList)
    )

    def _dissect(self, buf):
        self._init_triggerlist("rte_auth", buf[4:], self.__parse_auths)
        return len(buf)

    def __parse_auths(self, buf):
        off = 0
        auths = []

        while off + 20 <= len(buf):
            if buf[off: off + 2] == b"\xff\xff":
                auth_rte = Auth(buf[off: off + 20])
            else:
                auth_rte = RTE(buf[off: off + 20])
            # logger.debug("RIP: adding auth/rte: %s" % auth_rte)
            auths.append(auth_rte)
            off += 20
        return auths


class RTE(packetracer.Packet):
    __hdr__ = (
        ("family", "H", 2),
        ("route_tag", "H", 0),
        ("addr", "I", 0),
        ("subnet", "I", 0),
        ("next_hop", "I", 0),
        ("metric", "I", 1)
    )


class Auth(packetracer.Packet):
    __hdr__ = (
        ("rsvd", "H", 0xFFFF),
        ("type", "H", 2),
        ("auth", "16s", b"\x00" * 16)
    )



"""
Internet Control Message Protocol for IPv6.
https://tools.ietf.org/html/rfc2463
"""
import logging
import struct

from packetracer import checksum
from packetracer import packetracer
from packetracer import triggerlist
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

ICMP6_DST_UNREACH = 1  # dest unreachable, codes:
ICMP6_PACKET_TOO_BIG = 2  # packet too big
ICMP6_TIME_EXCEEDED = 3  # time exceeded, code:
ICMP6_PARAM_PROB = 4  # ip6 header bad

ICMP6_ECHO_REQUEST = 128  # echo service
ICMP6_ECHO_REPLY = 129  # echo reply
MLD_LISTENER_QUERY = 130  # multicast listener query
MLD_LISTENER_REPORT = 131  # multicast listener report
MLD_LISTENER_DONE = 132  # multicast listener done

# RFC2292 decls
ICMP6_MEMBERSHIP_QUERY = 130  # group membership query
ICMP6_MEMBERSHIP_REPORT = 131  # group membership report
ICMP6_MEMBERSHIP_REDUCTION = 132  # group membership termination

ND_ROUTER_SOLICIT = 133  # router solicitation
ND_ROUTER_ADVERT = 134  # router advertisment
ND_NEIGHBOR_SOLICIT = 135  # neighbor solicitation
ND_NEIGHBOR_ADVERT = 136  # neighbor advertisment
ND_REDIRECT = 137  # redirect

ICMP6_ROUTER_RENUMBERING = 138  # router renumbering

ICMP6_WRUREQUEST = 139  # who are you request
ICMP6_WRUREPLY = 140  # who are you reply
ICMP6_FQDN_QUERY = 139  # FQDN query
ICMP6_FQDN_REPLY = 140  # FQDN reply
ICMP6_NI_QUERY = 139  # node information request
ICMP6_NI_REPLY = 140  # node information reply

ICMP6_MAXTYPE = 201

pack_ipv6_icmp6 = struct.Struct(">16s16sII").pack


class ICMP6(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        ("code", "B", 0),
        ("sum", "H", 0)
    )

    def _calc_sum(self):
        try:
            # we need src/dst for checksum-calculation
            src, dst = self._lower_layer.src, self._lower_layer.dst
        except Exception:
            # not an IP packet as lower layer (src, dst not present) or invalid src/dst
            # logger.debug("could not calculate checksum: %r" % e)
            return

        # pseudoheader
        # packet length = length of upper layers
        self.sum = 0
        # logger.debug("TCP sum recalc: IP6= len(src)=%d\n%s\n%s\nhdr=%s\nbody=%s" %
        #			 (len(src), src, dst, self.header_bytes, self.body_bytes))
        pkt = self.header_bytes + self.body_bytes
        hdr = pack_ipv6_icmp6(src, dst, len(pkt), 58)
        # this will set the header status to changes, should be reset by calling bin()
        self.sum = checksum.in_cksum(hdr + pkt)

    # logger.debug(">>> new checksum: %0X" % self.sum)

    def _dissect(self, buf):
        self._init_handler(buf[0], buf[4:])
        return 4

    def _update_fields(self):
        try:
            if self.lower_layer._changed():
                self._calc_sum()
        except Exception:
            # no lower layer, nothing to update
            # logger.debug("%r" % ex)
            pass

    class Error(packetracer.Packet):
        __hdr__ = (("pad", "I", 0),)

    class Unreach(Error):
        pass

    class TooBig(Error):
        __hdr__ = (("mtu", "I", 1232),)

    class TimeExceed(Error):
        pass

    class ParamProb(Error):
        __hdr__ = (("ptr", "I", 0),)

    class Echo(packetracer.Packet):
        __hdr__ = (
            ("id", "H", 0),
            ("seq", "H", 0)
        )

    class ICMPv6Opt(packetracer.Packet):
        __hdr__ = (
            ("type", "B", 0),
            ("len", "B", 0)
        )

    @staticmethod
    def _parse_icmp6opt(buf):
        opts = []
        off = 0

        while off < len(buf):
            optlen = unpack_H(buf[1:3])[0] * 8
            opt = ICMP6.ICMPv6Opt(buf[off: off + optlen])
            opts.append(opt)
            off += optlen
        return opts

    class NeighbourSolicitation(packetracer.Packet):
        __hdr__ = (
            ("rsv", "4s", b"\x00" * 4),
            ("target", "16s", b"\x00" * 16),
            ("opts", None, triggerlist.TriggerList)
        )

        def _dissect(self, buf):
            self._init_triggerlist("opts", buf[20:], ICMP6._parse_icmp6opt)
            return len(buf)

    class NeighbourAdvertisement(packetracer.Packet):
        __hdr__ = (
            ("flags", "4s", b"\x00" * 4),
            ("target", "16s", b"\x00" * 16),
            ("opts", None, triggerlist.TriggerList)
        )

        def _dissect(self, buf):
            self._init_triggerlist("opts", buf[20:], ICMP6._parse_icmp6opt)
            return len(buf)

    __handler__ = {
        1: Unreach,
        2: TooBig,
        3: TimeExceed,
        4: ParamProb,
        128: Echo,
        129: Echo,
        135: NeighbourSolicitation,
        136: NeighbourAdvertisement
    }



"""Network Time Protocol v4"""
import logging

from packetracer import packetracer

logger = logging.getLogger("packetracer")

# Leap Indicator (LI) Codes
NO_WARNING = 0
LAST_MINUTE_61_SECONDS = 1
LAST_MINUTE_59_SECONDS = 2
ALARM_CONDITION = 3

# Mode Codes
RESERVED = 0
SYMMETRIC_ACTIVE = 1
SYMMETRIC_PASSIVE = 2
CLIENT = 3
SERVER = 4
BROADCAST = 5
CONTROL_MESSAGE = 6
PRIVATE = 7


class NTP(packetracer.Packet):
    __hdr__ = (
        ("flags", "B", 0x1C),  # li | v | mode
        ("stratum", "B", 0x2),
        ("interval", "B", 0x4),
        ("precision", "B", 0xE9),
        ("delay", "I", 0),
        ("dispersion", "I", 0),
        ("id", "4s", b"\x00\x01\x02\x03"),
        # timestamps: [seconds since 1.1.1900 | fraction of seconds]
        ("update_time", "8s", b"\x00" * 8),
        ("originate_time", "8s", b"" * 8),
        ("receive_time", "8s", b"" * 8),
        ("transmit_time", "8s", b"" * 8)
    )

    # li  v     mode [xx][xx x][xxx]
    def __get_v(self):
        return (self.flags >> 3) & 0x7

    def __set_v(self, value):
        self.flags = (self.flags & ~0x38) | ((value & 0x7) << 3)

    v = property(__get_v, __set_v)

    def __get_li(self):
        return (self.flags >> 6) & 0x3

    def __set_li(self, value):
        self.flags = (self.flags & ~0xC0) | ((value & 0x3) << 6)

    li = property(__get_li, __set_li)

    def __get_mode(self):
        return self.flags & 0x7

    def __set_mode(self, value):
        self.flags = (self.flags & ~0x7) | (value & 0x7)

    mode = property(__get_mode, __set_mode)



"""Remote Authentication Dial-In User Service."""

from packetracer import packetracer


# http://www.untruth.org/~josh/security/radius/radius-auth.html
# RFC 2865


class Radius(packetracer.Packet):
    __hdr__ = (
        ("code", "B", 0),
        ("id", "B", 0),
        ("len", "H", 4),
        ("auth", "16s", b"")
    )

    @staticmethod
    def parse_attrs(buf):
        """
        Parse attributes buffer into a list of (type, data) tuples.
        """
        attrs = []
        while buf:
            t = ord(buf[0])
            hlen = ord(buf[1])
            if hlen < 2:
                break
            d, buf = buf[2:hlen], buf[hlen:]
            attrs.append((t, d))
        return attrs


# Codes
RADIUS_ACCESS_REQUEST = 1
RADIUS_ACCESS_ACCEPT = 2
RADIUS_ACCESS_REJECT = 3
RADIUS_ACCT_REQUEST = 4
RADIUS_ACCT_RESPONSE = 5
RADIUS_ACCT_STATUS = 6
RADIUS_ACCESS_CHALLENGE = 11

# Attributes
RADIUS_USER_NAME = 1
RADIUS_USER_PASSWORD = 2
RADIUS_CHAP_PASSWORD = 3
RADIUS_NAS_IP_ADDR = 4
RADIUS_NAS_PORT = 5
RADIUS_SERVICE_TYPE = 6
RADIUS_FRAMED_PROTOCOL = 7
RADIUS_FRAMED_IP_ADDR = 8
RADIUS_FRAMED_IP_NETMASK = 9
RADIUS_FRAMED_ROUTING = 10
RADIUS_FILTER_ID = 11
RADIUS_FRAMED_MTU = 12
RADIUS_FRAMED_COMPRESSION = 13
RADIUS_LOGIN_IP_HOST = 14
RADIUS_LOGIN_SERVICE = 15
RADIUS_LOGIN_TCP_PORT = 16
# unassigned
RADIUS_REPLY_MESSAGE = 18
RADIUS_CALLBACK_NUMBER = 19
RADIUS_CALLBACK_ID = 20
# unassigned
RADIUS_FRAMED_ROUTE = 22
RADIUS_FRAMED_IPX_NETWORK = 23
RADIUS_STATE = 24
RADIUS_CLASS = 25
RADIUS_VENDOR_SPECIFIC = 26
RADIUS_SESSION_TIMEOUT = 27
RADIUS_IDLE_TIMEOUT = 28
RADIUS_TERMINATION_ACTION = 29
RADIUS_CALLED_STATION_ID = 30
RADIUS_CALLING_STATION_ID = 31
RADIUS_NAS_ID = 32
RADIUS_PROXY_STATE = 33
RADIUS_LOGIN_LAT_SERVICE = 34
RADIUS_LOGIN_LAT_NODE = 35
RADIUS_LOGIN_LAT_GROUP = 36
RADIUS_FRAMED_ATALK_LINK = 37
RADIUS_FRAMED_ATALK_NETWORK = 38
RADIUS_FRAMED_ATALK_ZONE = 39
# 40-59 reserved for accounting
RADIUS_CHAP_CHALLENGE = 60
RADIUS_NAS_PORT_TYPE = 61
RADIUS_PORT_LIMIT = 62
RADIUS_LOGIN_LAT_PORT = 63



"""
Linux cooked capture format
"""
import logging

from packetracer import packetracer
# handler
from packetracer.layer12 import can, arp, dtp, pppoe
from packetracer.layer3 import ip, ip6, ipx
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

# Ethernet payload types - http://standards.ieee.org/regauth/ethertype
LCC_TYPE_CAN = 0x000C  # CAN protocol
LCC_TYPE_PUP = 0x0200  # PUP protocol
LCC_TYPE_IP = 0x0800  # IPv4 protocol
LCC_TYPE_ARP = 0x0806  # address resolution protocol
LCC_TYPE_WOL = 0x0842  # Wake on LAN
LCC_TYPE_CDP = 0x2000  # Cisco Discovery Protocol
LCC_TYPE_DTP = 0x2004  # Cisco Dynamic Trunking Protocol
LCC_TYPE_REVARP = 0x8035  # reverse addr resolution protocol
LCC_TYPE_ETHTALK = 0x809B  # Apple Talk
LCC_TYPE_AARP = 0x80F3  # Appletalk Address Resolution Protocol
LCC_TYPE_8021Q = 0x8100  # IEEE 802.1Q VLAN tagging
LCC_TYPE_IPX = 0x8137  # Internetwork Packet Exchange
LCC_TYPE_NOV = 0x8138  # Novell
LCC_TYPE_IP6 = 0x86DD  # IPv6 protocol
LCC_TYPE_MPLS_UCAST = 0x8847  # MPLS unicast
LCC_TYPE_MPLS_MCAST = 0x8848  # MPLS multicast
LCC_TYPE_PPOE_DISC = 0x8863  # PPPoE Discovery
LCC_TYPE_PPOE_SESS = 0x8864  # PPPoE Session
LCC_TYPE_JUMBOF = 0x8870  # Jumbo Frames
LCC_TYPE_PROFINET = 0x8892  # Realtime-Ethernet PROFINET
LCC_TYPE_ATAOE = 0x88A2  # ATA other Ethernet
LCC_TYPE_ETHERCAT = 0x88A4  # Realtime-Ethernet Ethercat
LCC_TYPE_PBRIDGE = 0x88A8  # Provider Briding
LCC_TYPE_POWERLINK = 0x88AB  # Realtime Ethernet POWERLINK
LCC_TYPE_LLDP = 0x88CC  # Link Layer Discovery Protocol
LCC_TYPE_SERCOS = 0x88CD  # Realtime Ethernet SERCOS III
LCC_TYPE_FIBRE_ETH = 0x8906  # Fibre Channel over Ethernet
LCC_TYPE_FCOE = 0x8914  # FCoE Initialization Protocol (FIP)

PACKET_DIR_TO_US = 0
PACKET_DIR_FROM_US = 4


class LinuxCC(packetracer.Packet):
    __hdr__ = (
        ("dir", "H", 4),
        ("addrtype", "H", 0),
        ("addrlen", "H", 0),
        ("info", "Q", 0),
        ("type", "H", LCC_TYPE_IP)
    )

    __handler__ = {
        LCC_TYPE_CAN: can.CAN,
        LCC_TYPE_IP: ip.IP,
        LCC_TYPE_ARP: arp.ARP,
        LCC_TYPE_DTP: dtp.DTP,
        LCC_TYPE_IPX: ipx.IPX,
        LCC_TYPE_IP6: ip6.IP6,
        LCC_TYPE_PPOE_DISC: pppoe.PPPoE,
        LCC_TYPE_PPOE_SESS: pppoe.PPPoE
    }

    def _dissect(self, buf):
        htype = unpack_H(buf[14: 16])[0]
        # logger.debug("type: %X" % type)
        self._init_handler(htype, buf[16:])
        return 16



"""Trivial File Transfer Protocol (TFTP)"""
import logging
import re

from packetracer.packetracer import Packet
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

PROG_SPLIT_NULLBYTE = re.compile(b"\x00")
split_nullbyte = PROG_SPLIT_NULLBYTE.split

# Opcodes
OP_RRQ = 1  # read request
OP_WRQ = 2  # write request
OP_DATA = 3  # data packet
OP_ACK = 4  # acknowledgment
OP_ERR = 5  # error code

OPCODES_READ_WRITE = {OP_RRQ, OP_WRQ}
OPCODES_DATA_ACK = {OP_DATA, OP_ACK}

# Error codes
EUNDEF = 0  # not defined
ENOTFOUND = 1  # file not found
EACCESS = 2  # access violation
ENOSPACE = 3  # disk full or allocation exceeded
EBADOP = 4  # illegal TFTP operation
EBADID = 5  # unknown transfer ID
EEXISTS = 6  # file already exists
ENOUSER = 7  # no such user


class TFTP(Packet):
    __hdr__ = (
        ("opcode", "H", OP_RRQ),
        ("file", None, None),
        ("block", "H", 0),
        ("ttype", None, None)
    )

    def _dissect(self, buf):
        hlen = 2
        opcode = unpack_H(buf[: 2])
        # logger.debug("opcode: %d" % opcode)

        if opcode in OPCODES_DATA_ACK:
            pass
        elif opcode in OPCODES_READ_WRITE:
            self.block = None
            file, ttype = split_nullbyte(buf[2:], maxsplit=2)
            # logger.debug("file/ttype = %r / %r" % (file, ttype))
            self.file = file + b"\x00"
            self.ttype = ttype + b"\x00"
            hlen = 2 + len(self.file) + len(self.ttype)
        elif opcode == OP_ERR:
            pass
        return hlen



"""
Simple packet creation and parsing logic.
"""
import inspect
import logging
import random
import re
from ipaddress import IPv6Address, v6_int_to_packed
from struct import Struct

from packetracer.lazydict import LazyDict
# imported to make usable via import "packetracer.[FIELD_FLAG_AUTOUPDATE | FIELD_FLAG_IS_TYPEFIELD]"
from packetracer.packetracer_meta import MetaPacket
from packetracer.structcbs import pack_mac, unpack_mac, pack_ipv4, unpack_ipv4

logger = logging.getLogger("packetracer")
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARNING)

logger_streamhandler = logging.StreamHandler()
logger_formatter = logging.Formatter("%(levelname)s (%(funcName)s): %(message)s")
logger_streamhandler.setFormatter(logger_formatter)

logger.addHandler(logger_streamhandler)

PROG_NONVISIBLE_CHARS = re.compile(b"[^\x21-\x7e]")
HEADER_TYPES_SIMPLE = {int, bytes}

DIR_SAME = 1
DIR_REV = 2
DIR_UNKNOWN = 4
DIR_NOT_IMPLEMENTED = 255

ERROR_NONE = 0
ERROR_DISSECT = 1
ERROR_UNKNOWN_PROTO = 2
ERROR_NOT_UNPACKED = 4

VARFILTER_TYPES = {bytes, int}


class InvalidValuetypeException(Exception):
    pass


class Packet(object, metaclass=MetaPacket):
    """
    Base packet class, with metaclass magic to generate members from self.__hdr__ field.
    This class can be instatiated via:

        Packet(byte_string)
        Packet(key1=val1, key2=val2, ...)

    Every packet got a header and a body. Body-data can be raw byte string OR a packet itself
    (the body handler) which itself stores a packet etc. This continues until a packet only
    contains raw bytes (highest layer). The following schema illustrates the Packet-structure:

    Packet structure
    ================

    [Packet:
    headerfield_1
    headerfield_2
    ...
    headerfield_N
    [Body -> Packet:
        headerfield_1
        ...
        headerfield_N
        [Body: -> Packet:
            headerfields
            ...
            [Body: b"some_bytes"]
    ]]]

    A header definition like __hdr__ = (("name", "12s", b"defaultvalue"),) will define a header field
    having the name "name", format "12s" and default value b"defaultvalue" as bytestring. Fields will
    be added in order of definition. Extending classes should overwrite the "_dissect"-method in order to dissect
    given data.

    Requirements
    ============

    - Auto-decoding of headers via given format-patterns (defined via __hdr__)
    - Auto-decoding of body-handlers (IP -> parse IP-data -> add TCP-handler to IP -> parse TCP-data..)
    - Access of higher layers via layer1.higher_layer or "layer1[layerX]" notation
    - There are three types of headers:
    1) Simple constant fields (constant format)
        Format for __hdr__: ("name", "format", value [, FLAGS])

    2) Simple dynamic fields (byte string which changes in length)
        Format for __hdr__: ("name", None, b"bytestring" [, FLAGS])
        Such types MUST get initiated in _dissect() because there is no way in guessing
        the correct format when unpacking values!

    3) TriggerList (List containing Packets, bytes like b"xyz" or tuples like (ID, value))
        Format for __hdr__: ("name", None, TriggerList)

    The FLAGS value for simple constant and dynamic fields can be used to mark auto-update field
    (see packetracer_meta.py). This will create a variable XXX_au_active one time for a field XXX
    which can be used activate/deactivate the auto-update externally and which can be read in
    the bin()-method internally.
    - Convenient access for standard types (e.g. MAC, IP address) using string-representations
        This is done by appending "_s" to the attributename:
        ip.src_s = "127.0.0.1"
        ip_src_str = ip.src_s

        Implementation info:
        Convenient access should be set via varname_s = packetracer.Packet.get_property_XXX("varname")
        Get/set via is always done using strings (not byte strings).
    - Concatination via "layer1 + layer2 + layerX"
    - Header-values with length < 1 Byte should be set by using properties
    - Activate/deactivate non-TriggerList header fields by setting values (None=deactive, value=active)
    - Checksums (static auto fields in general) are auto-recalculated when calling
        bin(update_auto_fields=True) (default: active)
        The update-behaviour for every single field can be controlled via
        "pkt.VARNAME_au_active = [True|False]
    - Ability to check direction to other Packets via "[is_]direction()"
    - Access to next lower/upper layer
    - No correction of given raw packet-bytes e.g. checksums when creating a packet from it
        If the packet can't be parsed without correct data -> raise exception.
        The internal state will only be updated on changes to headers or data later on
    - General rule: less changes to headers/body-data = more performance

    Call-flows
    ==========

        packetracer(bytes)
            -> _dissect(): has to be overwritten, get to know/verify the real header-structure
                -> (optional): call _init_handler() initiating a handler representing an upper-layer
                -> (optional): call _init_triggerlist(name, b"bytes", dissect_callback)
                    to initiate a TriggerList field
                Note: headers won't be updated to buffer values until dissect finishes
            -> (optional) on access to simple headers: _unpack() sets all header values
            -> (optional) on access to TriggerList headers: lazy parsing gets triggered
            -> (optional) on access to body handler next upper layer gets initiated

        packetracer(keyword1=value, ...)
            -> (optional) set headers

        packetracer()
            -> sets standard values for simple headers

    """

    # Dict for saving "body type ids -> handler classes" globaly:
    # { class_name_current : {id_upper : handler_class_upper} }
    _id_handlerclass_dct = {}
    # Dict for saving "handler class -> body type ids" globaly:
    # { class_name_current : {handler_class_upper : id_upper} }
    _handlerclass_id_dct = {}
    # Constants for Packet-directions
    DIR_SAME = DIR_SAME
    DIR_REV = DIR_REV
    DIR_UNKNOWN = DIR_UNKNOWN
    DIR_NOT_IMPLEMENTED = DIR_NOT_IMPLEMENTED

    def __init__(self, *args, **kwargs):
        """
        Packet constructors:

        Packet(bytestring, target_class)
            Note: target_class is only meant for internal usage
        Packet(keyword1=val1, keyword2=val2, ...)

        bytestring -- packet bytes to build packet from, nonempty values are NOT allowed
        target_class -- For internal usage only: unpack until this class (meant eg for __getitem__(...))
        keywords -- keyword arguments correspond to header fields to be set
        """

        if args:
            if len(args) > 1:
                # Assume packet, target class given until which we unpack
                self._final_unpack_clz = args[1]._final_unpack_clz
            # Any Exception will be forwarded (SomePkt(bytes) or lazy dissect)
            # If this is the lowest layer the Exception has to be caught
            # logger.debug("dissecting: %s", self.__class__.__name__)
            header_len = self._dissect(args[0])
            # logger.debug("init header (+ body bytes): %s", self.__class__.__name__)
            # problem: len(args[0]) < _header_len -> can't be unpacked
            if len(args[0]) < header_len:
                raise Exception("Not enough bytes for packet class %s: given=%d < expected=%d" %
                                (self.__class__, len(args[0]), header_len))

            self._header_len = header_len
            self._header_cached = args[0][:header_len]

            if not self._body_changed:
                # _dissect(...) didn't call _init_handler(): set raw bytes.
                self._body_bytes = args[0][header_len:]
            # logger.warning("could not dissect in %s: %r" % (self.__class__.__name__, e))
            # Reset the changed-flags: original unpacked value = no changes
            self._reset_changed()
            self._unpacked = False
        else:
            if len(kwargs) > 0:
                # overwrite default parameters
                # logger.debug("new packet with keyword args (%s)", self.__class__.__name__)
                # _unpack is set to None: nothing to unpack until now

                for k, v in kwargs.items():
                    # logger.debug("setting: %s=%s", k, v)
                    setattr(self, k, v)
            # no reset: directly assigned = changed
            # keyword args means: allready unpacked (nothing to unpack)
            self._unpacked = True

    def _dissect(self, buf):
        """
        Dissect packet bytes by doing some (or nothing) of the following:
        - call self._init_triggerlist(...) to initiate TriggerLists
        - call self._init_handler(...) to initiate upper layer handler
        - activate/deactivate non-TriggerList fields by setting values/None to fields

        buf -- bytestring to be dissected
        return -- header length
        """
        # _dissect(...) was not overwritten: no changes to header, return original header length
        return self._header_len

    def __len__(self):
        """Return total length (= header + all upper layer data) in bytes."""
        if self._lazy_handler_data is not None:
            # lazy data present: avoid unneeded parsing
            # logger.debug("returning length from cached lazy handler in %s", self.__class__.__name__)
            return self.header_len + len(self._lazy_handler_data[1])
        elif self._higher_layer is not None:
            # logger.debug("returning length from present handler in %s, handler is: %s"\
            # % (self.__class__.__name__, self._higher_layer))
            return self.header_len + len(self._higher_layer)
        else:
            # Assume bodybytes are set
            # logger.debug("returning length from raw bytes in %s", self.__class__.__name__)
            return self.header_len + len(self._body_bytes)

    #
    # public access to header length: keep it uptodate
    #
    def _get_header_len(self):
        if self._header_changed and self._header_format_changed:
            # header has NOT changed if __init__ just finished -> avoid unneeded re-formating
            # update format to get the real length
            self._update_header_format()
        return self._header_len

    # update format if needed and return actual header size
    header_len = property(_get_header_len)

    def _get_dissect_error(self):
        return (self._errors & ERROR_DISSECT) != 0

    dissect_error = property(_get_dissect_error)
    errors = property(lambda obj: obj._errors)

    def is_error_present(self, error):
        """
        Check if one of packetracer.ERROR_XXX is present
        error -- the error to be check against internal error state
        """
        return (self._errors & error) != 0

    def _get_bodybytes(self):
        """
        Return raw data bytes or handler bytes (including all upper layers) if present.
        This is the same as calling bin() but excluding this header and without resetting changed-status.
        """
        if self._lazy_handler_data is not None:
            # no need to parse: raw bytes for all upper layers
            return self._lazy_handler_data[1]
        elif self._higher_layer is not None:
            # some handler was set
            hndl = self._higher_layer
            return hndl._pack_header() + hndl._get_bodybytes()
        else:
            # return raw bytes (no handler)
            return self._body_bytes

    def _set_bodybytes(self, value):
        """
        Set body bytes to value (bytestring). This will reset any handler.

        value -- a byte string (do NOT set to None)
        """
        if self._higher_layer is not None:
            # reset all handler data
            self._set_higherlayer(None)
        # logger.debug("setting new raw data: %s", value)
        self._body_bytes = value
        self._body_changed = True
        self._lazy_handler_data = None
        # logger.debug("notify after setting body bytes")
        self._notify_changelistener()

    # Get and set bytes for body. Note: this returns bytes even if higher_layer returns None.
    # Setting body_bytes will clear any handler (higher_layer will return None afterwards).
    body_bytes = property(_get_bodybytes, _set_bodybytes)

    def _get_higherlayer(self):
        """
        Retrieve next upper layer. This is the only direct way to do this.
        return -- handler object or None if not present.
        """
        if self._lazy_handler_data is not None:
            self._lazy_init_handler()
        return self._higher_layer

    @staticmethod
    def get_id_for_handlerclass(origin_class, handler_class):
        """
        return -- id associated for the given handler_class used in class origin_class.
            None if nothing was found. Example: origin_class = Ethernet, handler_class = IP,
            id will be ETH_TYPE_IP
        """
        try:
            return Packet._handlerclass_id_dct[origin_class][handler_class]
        except:
            # logger.debug("Could not find body handler id for %s in current class %s",
            # hndl.__class__, self.__class__)
            pass
        return None

    def _set_higherlayer(self, hndl):
        """
        Set body handler for this packet and make it accessible via layername[addedtypeclass]
        like ethernet[ip.IP]. If handler is None any handler will be reset and data will be set to an empty byte string.

        hndl -- the handler to be set: None or a Packet instance. Setting to None
            will clear any handler and set body_bytes to b"".
        """
        if self._higher_layer is not None:
            # clear old linked data of upper layer if body handler is already parsed
            # A.B -> A.higher_layer = x -> B.lower_layer = None
            # logger.debug("removing old data handler connections")
            self._higher_layer._lower_layer = None

        if hndl is None:
            # avoid (body_bytes=None, handler=None)
            self._body_bytes = b""
        else:
            # set a new body handler
            # associate ip, arp etc with handler-instance to call "ether.ip", "ip.tcp" etc
            self._body_bytes = None
            hndl._lower_layer = self

        self._higher_layer = hndl
        self._body_changed = True
        self._lazy_handler_data = None
        # logger.debug("notify after setting handler")
        self._notify_changelistener()

    # deprecated, wording "higher_layer/highest_layer layer is more consistent
    upper_layer = property(_get_higherlayer, _set_higherlayer)
    # Get/set body handler. Note: this will force lazy dissecting when reading
    higher_layer = property(_get_higherlayer, _set_higherlayer)

    def _set_lower_layer(self, val):
        if self._lower_layer is not None:
            # remove upper layer (us) from current lower layer before
            # setting a new lower layer
            self._lower_layer.higher_layer = None

        self._lower_layer = val

    # Get/set next lower body handler
    lower_layer = property(lambda pkt: pkt._lower_layer, _set_lower_layer)

    def _lowest_layer(self):
        current = self

        while current._lower_layer is not None:
            current = current._lower_layer

        return current

    def _get_highest_layer(self):
        current = self

        # unpack all layer, assuming string class will be never found
        self._final_unpack_clz = str.__class__

        while current.higher_layer is not None:
            current = current.higher_layer

        return current

    def _set_highest_layer(self, layer):
        """
        Replaces the current highest layer with a new one,
        eg with new layer "D" A.B.C becomes A.B.D.
        """
        layer_to_change = self.highest_layer.lower_layer

        if layer_to_change is None:
            return

        layer_to_change.higher_layer = layer

    # get lowest layer
    lowest_layer = property(_lowest_layer)
    # get top layer
    highest_layer = property(_get_highest_layer)

    def disconnect_layer(self):
        """
        Disconnect and return this layer. Connects lower and upper layer
        with each other. This is the same as 'pkt.lower_layer = pkt.higher_layer'
        without returning the middle layer (pkt).
        return -- This layer
        """
        # Connect lower/upper layer of this layer
        if self.lower_layer is not None and self.higher_layer is not None:
            self.lower_layer.higher_layer = self.higher_layer

        self.lower_layer = None
        self.higher_layer = None

        return self

    def _lazy_init_handler(self):
        """
        Lazy initialize the handler previously set by _init_handler.
        Make sure this is not called more than once
        """
        handler_data = self._lazy_handler_data

        try:
            # Instantiate handler class using lazy data buffer
            # See _init_handler() for 2nd place where handler instantation takes place
            # logger.debug("lazy parsing using: %s", handler_data)
            type_instance = handler_data[0](handler_data[1], self)

            self._set_higherlayer(type_instance)
            # this was a lazy init: same as direct dissecting -> no body change
            self._body_changed = False
        except:
            # error on lazy dissecting: set raw bytes
            self._errors |= ERROR_DISSECT
            self._body_bytes = handler_data[1]
        # logger.warning("Can't set handler data (malformed packet?): base=%s handler_type/handlerclass=%r, reason: %s",
        #	self.__class__, handler_data[0], ex)
        self._lazy_handler_data = None

    def __getitem__(self, packet_type):
        """
        Check every layer upwards (inclusive this layer) for the given Packet class
        and return the first matched instance or None if nothing was found.

        packet_type -- Packet class to search for like TCP or combined
            multi-value like Ethernet,IP,TCP. For multi-value the last
            given Type in sequence will be returned.
        return -- First finding of packet_type or None if nothing was found
        """
        p_instance = self

        # multi-value index search
        if type(packet_type) is tuple:
            type_cnt = 0
            packet_type_len = len(packet_type)

            for pkt_clz in packet_type:
                if pkt_clz != p_instance.__class__:
                    # mismatch
                    return None

                type_cnt += 1
                # highest layer reached
                if p_instance.higher_layer is None:
                    break
                elif type_cnt != packet_type_len:
                    # end of match sequence in packet_type not reached, go higher
                    p_instance = p_instance.higher_layer

            # return last matching layer
            return p_instance if type_cnt == packet_type_len else None
        # single-value index search
        else:
            # set most top layer to be unpacked, __getattr__() could be called unpacking lazy data
            self._final_unpack_clz = packet_type

            while not type(p_instance) is packet_type:
                # this will auto-parse lazy handler data via _get_higherlayer()
                p_instance = p_instance.higher_layer

                if p_instance is None:
                    break

            # logger.debug("returning found packet-handler: %s->%s", type(self), type(p_instance))
            return p_instance

    def __iter__(self):
        """
        Iterate over every layer starting from first layer.
        To start from the lowest layer use "for l in pkt.lowest_layer".
        """
        p_instance = self
        # Unpack until highest layer; assume string class never gets found as layer
        self._final_unpack_clz = str.__class__

        while p_instance is not None:
            yield p_instance
            # this will auto-parse lazy handler data via _get_higherlayer()
            p_instance = p_instance.higher_layer

            if p_instance is None:
                break

    def __contains__(self, clz):
        return self[clz] is not None

    def __eq__(self, clz):
        """
        Compare class of this object to the given class/object
        """
        # convert object to its class
        if not type(clz) == MetaPacket:
            clz = clz.__class__
        return self.__class__ == clz

    def dissect_full(self):
        """
        Recursive unpack ALL data inlcuding lazy header etc up to highest layer inlcuding danymic fields.
        """
        for name in self._header_field_names:
            self.__getattribute__(name)

        try:
            self.higher_layer.dissect_full()
        except:
            # no handler present
            pass

    def __add__(self, packet_or_bytes_to_add):
        """
        Concatinate a packet with another packet or bytes.
        Note: Beware of side effects as Packets remain connected until removed, eg via pkt.higher_layer = None.

        packet_or_bytes_to_add -- The packet or bytes to be added as highest layer
        """
        if type(packet_or_bytes_to_add) is not bytes:
            self.highest_layer.higher_layer = packet_or_bytes_to_add
        else:
            self.highest_layer.body_bytes += packet_or_bytes_to_add
        return self

    def __iadd__(self, packet_or_bytes_to_add):
        """
        Concatinate a packet with another packet or bytes.
        Note: Beware of side effects as Packets remain connected

        packet_or_bytes_to_add -- The packet or bytes to be added as highest layer
        """
        if type(packet_or_bytes_to_add) is not bytes:
            self.highest_layer.higher_layer = packet_or_bytes_to_add
        else:
            self.highest_layer.body_bytes += packet_or_bytes_to_add
        return self

    def split_layers(self):
        """
        Splits all layers to indepent ones starting from this one not connectedto each other
        e.g. A.B.C -> [A, B, C]
        return -- [layer1, layer2, ...]
        """
        layers = [layer for layer in self]

        for layer in layers:
            # avoid overwriting bytes, only reset handler
            if layer._body_bytes is None:
                layer.higher_layer = None
            layer.lower_layer = None
        return layers

    def _summarize(self):
        """
        Print a summary of this layer state. Shows all header, even deactivated ones.
        """
        # values need to be unpacked to be shown
        if not self._unpacked:
            self._unpack()

        # create key=value descriptions
        # show all header even deactivated ones
        layer_sums_l = []

        for name in self._header_field_names:
            name_real = name[1:]
            val = getattr(self, name_real)
            value_alt = ""
            value_translated = ""

            try:
                # Try to get convenient name
                value_alt = " = " + getattr(self, name_real + "_s")
            except:
                # Not set
                pass

            try:
                # Try to get translated name
                # logger.debug("Looking for translation: %s -> %s" % (self.__class__, name_real + "_t"))
                value_translated = getattr(self, name_real + "_t")

                if value_translated != "":
                    # Nothing found
                    value_translated = " = " + value_translated
            except:
                # Not set
                pass

            # Values: int
            if type(val) is int:
                format = getattr(self, name + "_format")
                layer_sums_l.append("%-12s (%s): 0x%X = %d = %s" % (name_real, format, val, val,
                                                                    bin(val)) + value_alt + value_translated)
            # Inactive
            elif val is None:
                layer_sums_l.append("%-16s: (inactive)" % name_real)
            else:
                # Values: bytes
                if type(val) is bytes:
                    bts_cnt = "(%d)" % len(val)
                    layer_sums_l.append("%-9s %6s: %s" % (name_real, bts_cnt, val) +
                                        value_alt + value_translated)
                # Values Triggerlist (can contain Packets, tuples, bytes)
                else:
                    layer_sums_l.append("%-16s: %s" % (name_real, val) + value_alt)

        if self.higher_layer is None:
            # No upper layer present: describe body bytes
            bts_cnt = "(%d)" % len(self.body_bytes)
            layer_sums_l.append("%-9s %6s: " % ("bodybytes", bts_cnt) + "%s" % self.body_bytes)

        layer_sums = "%s\n\t%s" % (
            self.__module__[9:] + "." + self.__class__.__name__,
            "\n\t".join(layer_sums_l))

        return layer_sums

    def __str__(self):
        # recalculate fields like checksums, lengths etc
        if self._header_changed or self._body_changed:
            # logger.debug("header/body changed: need to reparse (%s)", self.__class__)
            self.bin()
        # this does lazy init of handler
        upperlayer_str = "\n%s" % self.higher_layer if self.higher_layer is not None else ""
        return self._summarize() + upperlayer_str

    def _unpack(self):
        """
        Unpack a full layer (set field values) unpacked (extracted) from cached header bytes.
        This will use the current value of _header_cached to set all field values.
        NOTE:
        - This is only called by the Packet class itself
        - This is called prior to changing ANY header values
        """
        # Needed to set here (and not at the end) to avoid recursive calls
        self._unpacked = True
        # logger.debug("unpacking header: %s", self._header_field_names)
        # we need the whole format when:
        # format changed or some TriggestLists are non-empty (not yet dissected)
        if self._header_format_changed:
            self._update_header_format()

        # logger.debug("unpacking 1: %s, %s,\n%s,\n(format via xxx_format) %s,\n%s,\n%s\nformat.size %d\ncached size: %d" %
        #	(self.__class__,
        #	self._header_field_names,
        #	self._header_format.format,
        #	[self_getattr(name + "_format") for name in self._header_field_names],
        #	[self_getattr(name + "_active") for name in self._header_field_names],
        #	self._header_cached,
        #	self._header_format.size,
        #	len(self._header_cached)))

        # logger.debug([self_getattr(name) for name in self._header_field_names])
        try:
            header_unpacked = self._header_format.unpack(self._header_cached)
        except:
            self._errors |= ERROR_NOT_UNPACKED
            # just warn user about incomplete data
            errormsg = "could not unpack in: %s, format: %s, names: %s, value to unpack: %s (%d bytes)," \
                       " invalid value for format?" % (
                           self.__class__.__name__,
                           self._header_format.format,
                           self._header_field_names,
                           self._header_cached,
                           len(self._header_cached)
                       )

            logger.warning(errormsg)
            return
        # logger.debug("unpacking via format: %s -> %s", self._header_format.format, header_unpacked)
        cnt = 0
        self_setattr = self.__setattr__
        self_getattr = self.__getattribute__

        # logger.debug("unpacking 2: %s, %s -> %s,\n%s,\n %s\n",
        #	(self.__class__, header_unpacked, self._header_field_names,
        # 	[self_getattr(name + "_format") for name in self._header_field_names],
        # 	[self_getattr(name + "_active") for name in self._header_field_names])
        for name in self._header_field_names:
            # only set values if active simple field
            if self_getattr(name + "_active"):
                if self_getattr(name + "_format") is not None:
                    # logger.debug("!!!!! unpacking value: %s -> %s", name, header_unpacked[cnt])
                    self_setattr(name, header_unpacked[cnt])
                # inactive fields are not in unpacked list
                cnt += 1

    def reverse_address(self):
        """
        Reverse source <-> destination address of THIS packet. This is at minimum
        defined for: Ethernet, IP, TCP, UDP
        """
        pass

    def reverse_all_address(self):
        """
        Reverse source <-> destination address of EVERY packet upwards including this one
        (reverse_address has to be implemented).
        """
        current_hndl = self

        while current_hndl is not None:
            current_hndl.reverse_address()
            current_hndl = current_hndl.higher_layer

    def _init_handler(self, hndl_type, buffer):
        """
        Called by overwritten "_dissect()". 1) Store handler for later dissect or 2) directly dissect it.
        Notes for 2): Initiate the handler-parser using the given buffer and set it using _set_higherlayer()
        On any error this will set raw bytes given for body data.

        hndl_type -- A value to place the handler in the handler-dict like
            dict[Class.__name__][hndl_type] (eg type-id, port-number)
        buffer -- The buffer to be used to create the handler
        """
        # Empty buffer must lead to empty body. Initiating packets using empty buffer
        # would lead to wrong (default) values
        if len(buffer) == 0:
            # logger.debug("empty buffer given for _init_handler()!")
            return

        # self.__class__ MUST be contained, otherwise calling _init_handler() would be illegal
        try:
            # Likely to succeed
            clz = Packet._id_handlerclass_dct[self.__class__][hndl_type]
        except:
            self.body_bytes = buffer
            self._errors |= ERROR_UNKNOWN_PROTO
            # errormsg = "Unknown upper layer type for %s: %d, feel free to implement" % (
            #	self.__class__, hndl_type)
            # logger.warning(errormsg)
            return

        try:
            if self._final_unpack_clz is None or self._final_unpack_clz is self.__class__:
                # set lazy handler data, __getattr__() will be called on access
                # to handler (field not yet initiated)
                # logger.debug("setting handler name: %s -> %s", self.__class__.__name__, clz_name)
                self._lazy_handler_data = [clz, buffer]
                # set name although we don't set a handler (needed for direction() et al)
                self._body_bytes = None
                # avoid setting body_bytes by _unpack()
                self._body_changed = True
            else:
                # Continue parsing next upper layer, happens on "__iter__()": avoid unneeded lazy-data
                # handling/creating uneeded meta data for later body handling
                # logger.debug("--------> direct init in: %s", self.__class__.__name__)
                type_instance = clz(buffer, self)
                self._set_higherlayer(type_instance)
        except:
            # logger.warning("Can't set handler data (malformed?): base=%s handler_type/id=%r, reason: %s",
            #	self.__class__, Packet._id_handlerclass_dct[self.__class__][hndl_type], ex)
            # set raw bytes as data (eg handler class not found)
            self.body_bytes = buffer
            self._errors |= ERROR_DISSECT

    def _init_triggerlist(self, name, bts, dissect_callback):
        """
        Inititiate a TriggerList field. It will be dissected ondemand.

        name -- name of the dynamic filed to be initiated
        bts -- bts to be dissected
        dissect_callback -- callback to be used to dissect, signature:
            callback(bytes) -> returns list of bytes, packets, ...
        """
        self.__setattr__("_%s" % name, [bts, dissect_callback])
        self._header_format_changed = True

    def direction_all(self, other_packet):
        """
        Check for direction on ALL layers from this one upwards.
        This continues upwards until no body handler can be found anymore.
        The extending class can overwrite direction() to implement an individual check,

        other_packet -- Packet to be compared with this Packet
        return -- Bitwise AND-concatination of all directions of ALL layers starting from
            this one upwards. Directions are: [DIR_SAME | DIR_REV | DIR_UNKNOWN].
            This can be checked via eg "direction_found & DIR_SAME"
        """
        dir_ext = self.direction(other_packet)
        # logger.debug("direction of %s: %d", self.__class__, dir_ext)

        try:
            # check upper layers and combine current result
            # logger.debug("direction? checking next layer")
            dir_upper = self.higher_layer.direction_all(other_packet.higher_layer)

            return dir_ext & dir_upper
        except AttributeError:
            # one of both _higher_layer was None
            # Example: TCP ACK (last step of handshake, no payload) <-> TCP ACK + Telnet
            # logger.debug("AttributeError, direction: %d", dir_ext)
            # logger.debug(e)
            return dir_ext

    def direction(self, other):
        """
        Check if this layer got a specific direction compared to "other". Can be overwritten.

        return -- [DIR_SAME | DIR_REV | DIR_UNKNOWN | DIR_NOT_IMPLEMENTED]
        """
        return Packet.DIR_NOT_IMPLEMENTED

    def is_direction(self, packet2, direction):
        """
        Same as "direction_all()" but using explicit direction to be checked.
        As direction_all can be DIR_SAME and DIR_REV at the same time, this call
        is more clearly.

        packet2 -- packet to be compared to this packet
        direction -- check for this direction (DIR_...)
        return -- True if direction is found in this packet, False otherwise.
        """
        # logger.debug("direction_all & direction = %d & %d", self.direction_all(packet2), direction)
        return self.direction_all(packet2) & direction == direction

    def _update_higherlayer_id(self):
        """
        Updates the upperlayer id named by _id_fieldname (FIELD_FLAG_IS_TYPEFIELD was
        set) based on the upperlayer class and simply assigning the associated id to that field.

        Example: current layer = Ethernet, id field = type, body handler class = IP, eth.type
        will be set to ETH_TYPE_IP.

        If updating the type id is more complex than a simple assignmet this method has to
        be overwritten.
        """
        # do nothing if:
        # type id field not known or this is a parsed packet (non self-made) or we got no body handler
        # or nothing has changed
        # logger.debug("%s -> _id_fieldname: %s", self.__class__, self._id_fieldname)
        if self._id_fieldname is None \
                or not self._body_changed \
                or self._higher_layer is None \
                or not self.__getattribute__("%s_au_active" % self._id_fieldname) \
                or self._lazy_handler_data is not None:
            # logger.debug("Will NOT update!")
            return

        # logger.debug("will update handler id, %s / %s / %s / %s",
        #	self._id_fieldname,
        #	self.__getattribute__("%s_au_active" % self._id_fieldname),
        #	self._lazy_handler_data,
        #	self._body_changed)
        try:
            handler_clz = self._higher_layer.__class__
            # logger.debug("handler class is: %s", handler_clz)

            self.__setattr__(self._id_fieldname,
                             Packet._handlerclass_id_dct[self.__class__][handler_clz])
        except:
            # No type id found, something like eth + Telnet
            # logger.debug("no type id found for %s, class: %s -> %s" %
            #	(self._higher_layer.__class__, self.__class__, handler_clz))
            pass

    def _update_fields(self):
        """
        Overwrite this to update header fields. Only gets called if this or any other upper layer has changed.
        Callflow on a packet "pkt = layer1 + layer2 + layer3 -> pkt.bin()":
        layer3._update_fields() -> layer2._update_fields() -> layer1._update_fields() ...
        """
        pass

    def bin(self, update_auto_fields=True):
        """
        Return this header and body (including all upper layers) as byte string
        and reset changed-status.

        update_auto_fields -- if True auto-update fields like checksums, else leave them be
        """
        # Update all above layers until a non-handler layer is found
        if update_auto_fields and self._changed():
            # Collect layers to be updated:
            # Iterate update for A.B.C like C->B->A: A needs uptodate B and C,
            # B needs uptodate C
            layers = []
            layer_it = self

            while layer_it is not None:
                layers.append(layer_it)
                # Upper layer is not yet dissected but could need update:
                # ip:changed + tcp:notchanged/parsed -> tcp needs update
                if layer_it._lazy_handler_data is not None:
                    # Next upper layer forces update in layet_it
                    if layer_it._header_changed and \
                            layer_it._lazy_handler_data[0].__class__ in layer_it._update_dependants:
                        # Force dissecting
                        layer_it = layer_it.higher_layer
                    else:
                        layer_it = None
                else:
                    layer_it = layer_it.higher_layer
            # Start from the top
            layers.reverse()

            for layer in layers:
                layer._update_fields()

        if self._lazy_handler_data is not None:
            # logger.debug("Got lazy data layer: %s -> %s", self.__class__, self._lazy_handler_data[0])
            # no need to parse, just take lazy handler data bytes
            bodybytes_tmp = self._lazy_handler_data[1]
        elif self._higher_layer is not None:
            # logger.debug("Got upper layer: %s -> %s", self.__class__, self._higher_layer.__class__)
            # Don't update fields, this was already done above
            bodybytes_tmp = self.higher_layer.bin(update_auto_fields=False)
        else:
            # logger.debug("Got raw bytes in %s", self.__class__)
            # raw bytes
            bodybytes_tmp = self._body_bytes

        header_tmp = self._pack_header()

        # now every layer got informed about our status, reset it
        self._reset_changed()
        return header_tmp + bodybytes_tmp

    def _update_header_format(self):
        """
        Update format of this packet header. Needs to be called on changes to dynamic fields.
        """
        header_format = [">"]
        header_format_append = header_format.append
        self_getattr = self.__getattribute__

        for name in self._header_field_names:
            if not self_getattr(name + "_active"):
                continue

            val = self_getattr(name)

            if val.__class__ in HEADER_TYPES_SIMPLE:  # assume bytes/int
                header_format_append(self_getattr(name + "_format"))
            # logger.debug("adding format for (simple): %s, %s, val: %s format: %s",
            # self.__class__, name, self_getattr(name), self_getattr(name + "_format"))
            else:  # assume TriggerList
                if val.__class__ == list:
                    # TriggerList not yet initiated: take cached value
                    header_format_append("%ds" % len(val[0]))
                # logger.debug("adding format for: %s, %s, val: %s", self.__class__, name, val[0])
                else:
                    # Assume packet
                    header_format_append("%ds" % len(val.bin()))

        self._header_format = Struct("".join(header_format))
        self._header_len = self._header_format.size
        self._header_format_changed = False

    def _pack_header(self):
        """
        Return header as byte string.
        """
        if not self._header_changed:
            # return cached data if nothing changed
            # logger.warning("returning cached header (hdr changed=%s): %s->%s",
            # self._header_changed, self.__class__.__name__, self._header_cached)
            return self._header_cached

        if not self._unpacked:
            # this happens on: Packet(b"bytes") -> only changes to TriggerList. We need to unpack buffer values
            # to re-read them for header packing
            self._unpack()
        elif self._header_format_changed:
            # _unpack will call _update_header_format() if needed
            # real format needed for correct packing
            self._update_header_format()

        self_getattr = self.__getattribute__
        header_values = []
        header_values_append = header_values.append

        for name in self._header_field_names:
            if not self_getattr(name + "_active"):
                continue
            val = self_getattr(name)
            # two options:
            # - simple type (int, bytes, ...)	-> add given value
            # - TriggerList	(found via format None) -> call bin()
            if val.__class__ in HEADER_TYPES_SIMPLE:  # assume bytes/int
                header_values_append(val)
            else:  # assume TriggerList
                if val.__class__ == list:
                    header_values_append(val[0])
                else:
                    # Assume packet
                    header_values_append(val.bin())

        # logger.debug("header bytes for %s: %s = %s",
        # 	self.__class__.__name__, self._header_format.format, header_bytes)
        # Individual unpacking is about 4 times slower than cumulative
        self._header_cached = self._header_format.pack(*header_values)
        # logger.debug(">>> cached header: %s (%d)", self._header_cached, len(self._header_cached))
        self._header_changed = False

        return self._header_cached

    # readonly access to header
    header_bytes = property(_pack_header)

    def _changed(self):
        """
        Check if this or any upper layer changed in header or body

        return -- True if header or body changed, else False
        """
        changed = False
        p_instance = self

        while p_instance is not None:
            if p_instance._header_changed or p_instance._body_changed:
                changed = True
                break
            elif p_instance._lazy_handler_data is None:
                # one layer up, stop if next layer is not yet initiated which means: no change
                p_instance = p_instance.higher_layer
            else:
                # nothing changed upwards: lazy handler data still present/nothing got parsed
                break
        return changed

    def _reset_changed(self):
        """Set the header/body changed-flag to False. This won't clear caches."""
        self._header_changed = False
        self._body_changed = False

    def _add_change_listener(self, listener_cb):
        """
        Add a new callback to be called on changes to header or body.

        listener_cb -- the change listener to be added as callback-function
        """
        try:
            self._changelistener.add(listener_cb)
        except:
            # change listener not yet initiated
            self._changelistener = {listener_cb}

    def _remove_change_listener(self):
        """
        Remove all change listener.
        """
        if self._changelistener is not None:
            self._changelistener.clear()

    def _notify_changelistener(self):
        """
        Notify listener about changes in header or body using signature callback(self).
        This is primarily meant for triggerlist to react
        on changes in packets like Triggerlist[packet1, packet2, ...].
        """
        # logger.debug("packet is notifying!!!")
        # no listener added so far -> nothing to notify
        if self._changelistener is None:
            return

        for listener_cb in self._changelistener:
            # logger.debug("notify...")
            listener_cb()

    @classmethod
    def load_handler(cls, clz_add, handler):
        """
        Load Packet handler classes using a shared dictionary.

        clz_add -- class for which handler has to be added
        handler -- dict of handlers to be set like { id | (id1, id2, ...) : class }, id can be a tuple of values
        """
        if clz_add in Packet._id_handlerclass_dct:
            # logger.debug("handler already loaded: %s", clz_name)
            return

        # logger.debug("adding classes as handler: [%s] = %s", clz_add, handler)

        Packet._id_handlerclass_dct[clz_add] = {}
        Packet._handlerclass_id_dct[clz_add] = {}

        for handler_id, packetclass in handler.items():
            # packetracer.Packet.load_handler(IP, { ID : class } )
            if type(handler_id) is not tuple:
                Packet._id_handlerclass_dct[clz_add][handler_id] = packetclass
                Packet._handlerclass_id_dct[clz_add][packetclass] = handler_id
            else:
                # logger.debug("loading multi-it handler: clz_add=%s, packetclass=%s, handler_id[0]=%s" %
                #	(clz_add, packetclass, handler_id[0]))
                # packetracer.Packet.load_handler(IP, { (ID1, ID2, ...) : class } )
                for id_x in handler_id:
                    Packet._id_handlerclass_dct[clz_add][id_x] = packetclass
                # ambiguous relation of "handler class -> type ids", take 1st one
                Packet._handlerclass_id_dct[clz_add][packetclass] = handler_id[0]

    def hexdump(self, length=16, only_header=False):
        """
        length -- amount of bytes per line
        only_header -- if True: just dump header, else header + body (default)

        return -- hexdump output string for this packet (header or header + body).
        """
        bytepos = 0
        res = []

        if only_header:
            buf = self._pack_header()
        else:
            buf = self.bin()
        buflen = len(buf)

        while bytepos < buflen:
            line = buf[bytepos: bytepos + length]
            hexa = " ".join(["%02x" % x for x in line])
            # line = line.translate(__vis_filter)
            line = re.sub(PROG_NONVISIBLE_CHARS, b".", line)
            res.append("  %04d:      %-*s %s" % (bytepos, length * 3, hexa, line))
            bytepos += length
        return "\n".join(res)


#
# utility functions
# These could be put into separate modules but this would lead to recursive import problems.
#
# avoid unneeded references for performance reasons
randint = random.randint


def byte2hex(buf):
    """Convert a bytestring to a hex-represenation:
    b'1234' -> '\x31\x32\x33\x34'"""
    return "\\x" + "\\x".join(["%02X" % x for x in buf])


# MAC address
def mac_str_to_bytes(mac_str):
    """Convert mac address AA:BB:CC:DD:EE:FF to byte representation."""
    return b"".join([bytes.fromhex(x) for x in mac_str.split(":")])


def mac_bytes_to_str(mac_bytes):
    """Convert mac address from byte representation to AA:BB:CC:DD:EE:FF."""
    return "%02X:%02X:%02X:%02X:%02X:%02X" % unpack_mac(mac_bytes)


def get_rnd_mac():
    """Create random mac address as bytestring"""
    return pack_mac(randint(0, 255), randint(0, 255), randint(0, 255),
                    randint(0, 255), randint(0, 255), randint(0, 255))


def get_property_mac(varname):
    """Create a get/set-property for a MAC address as string-representation."""
    # logger.debug("--------------------- returning property")
    return property(
        lambda obj: mac_bytes_to_str(obj.__getattribute__(varname)),
        lambda obj, val: obj.__setattr__(varname, mac_str_to_bytes(val))
    )


# IPv4 address
def ip4_str_to_bytes(ip_str):
    """Convert ip address 127.0.0.1 to byte representation."""
    ips = [int(x) for x in ip_str.split(".")]
    return pack_ipv4(ips[0], ips[1], ips[2], ips[3])


def ip4_bytes_to_str(ip_bytes):
    """Convert ip address from byte representation to 127.0.0.1."""
    return "%d.%d.%d.%d" % unpack_ipv4(ip_bytes)


def get_rnd_ipv4():
    """Create random ipv4 adress as bytestring"""
    return pack_ipv4(randint(0, 255), randint(0, 255), randint(0, 255), randint(0, 255))


def get_property_ip4(var):
    """Create a get/set-property for an IP4 address as string-representation."""
    return property(
        lambda obj: ip4_bytes_to_str(obj.__getattribute__(var)),
        lambda obj, val: obj.__setattr__(var, ip4_str_to_bytes(val))
    )


# IPv6 address
def ip6_str_to_bytes(ip6_str):
    """Convert ip address 127.0.0.1 to byte representation."""
    return v6_int_to_packed(int(IPv6Address(ip6_str)))


def ip6_bytes_to_str(ip6_bytes):
    """Convert ip address from byte representation to 127.0.0.1."""
    return str(IPv6Address(ip6_bytes))


def get_property_ip6(var):
    """Create a get/set-property for an IP6 address as string-representation."""
    return property(
        lambda obj: ip6_bytes_to_str(obj.__getattribute__(var)),
        lambda obj, val: obj.__setattr__(var, ip6_str_to_bytes(val))
    )


# DNS names
def dns_name_decode(name, cb_mc_bytes=lambda: b""):
    """
    DNS domain name decoder (bytes to string)

    name -- example: b"\x03www\x07example\x03com\x00"
    cb_bytes -- callback to get bytes used to find name in case of Message Compression
        cb_bytes_pointer(): bytes
    return -- example: "www.example.com."
    """
    # ["www", "example", "com"]
    name_decoded = []
    parsed_pointers = set()
    off = 1
    buf = name

    while off < len(buf):
        size = buf[off - 1]
        if size == 0:
            break
        elif (size & 0b11000000) == 0:
            # b"xxx" -> "xxx"
            name_decoded.append(buf[off:off + size].decode())
            off += size + 1
        else:
            # DNS message compression
            off = (((buf[off - 1] & 0b00111111) << 8) | buf[off]) + 1
            buf = cb_mc_bytes()
            # logger.debug("Got compressed DNS: %s" % buf)

            if off in parsed_pointers:
                # dns message loop, abort...
                break
            parsed_pointers.add(off)
    return ".".join(name_decoded) + "."


def dns_name_encode(name):
    """
    DNS domain name encoder (string to bytes)

    name -- example: "www.example.com"
    return -- example: b'\x03www\x07example\x03com\x00'
    """
    name_encoded = [b""]
    # "www" -> b"www"
    labels = [part.encode() for part in name.split(".") if len(part) != 0]

    for label in labels:
        # b"www" -> "\x03www"
        name_encoded.append(chr(len(label)).encode() + label)
    return b"".join(name_encoded) + b"\x00"


def get_property_dnsname(var, cb_mc_bytes=lambda obj: b""):
    """
    Create a get/set-property for a DNS name.

    cb_bytes -- callback to get bytes used to find name in case of Message Compression
        cb_bytes_pointer(containing_obj) -- bytes
    """
    return property(
        lambda obj: dns_name_decode(obj.__getattribute__(var),
                                    cb_mc_bytes=lambda: cb_mc_bytes(obj)),
        lambda obj, val: obj.__setattr__(var, dns_name_encode(val))
    )


def get_property_bytes_num(var, format_target):
    """
    Creates a get/set-property for "bytes (format Xs) <-> number" where len(bytes) is not 2**x.
    Sometimes numbers aren't encoded as multiple of 2 (see SSL -> Handshake -> 3 bytes = integer???).
    That's bad. How to convert between both representations? Well...

    var -- varname to create a property for
    format_target -- real format of the theader used to create a number.

    Note: only use with static headers
    """
    format_target_struct = Struct(format_target)
    format_target_unpack = format_target_struct.unpack
    format_target_pack = format_target_struct.pack
    format_varname_s = ("_%s" % var) + "_format"

    def get_formatlen_of_var(obj):
        format_var_s = obj.__getattribute__(format_varname_s)

        if format_var_s is None:
            logger.warning("Got None format for %s, can't convert for convenience!", var)
            return 0

        return Struct(format_var_s).size

    def get_val_bts_to_int(obj):
        format_var_len = get_formatlen_of_var(obj)
        prefix_bts = (b"\x00" * (format_target_struct.size - format_var_len))
        return format_target_unpack(prefix_bts + obj.__getattribute__(var))[0]

    def set_val_int_to_bts(obj, val):
        format_var_len = get_formatlen_of_var(obj)
        obj.__setattr__(var, format_target_pack(format_target, val)[:-format_var_len])

    return property(
        # bytes -> int
        get_val_bts_to_int,
        # int -> bytes
        set_val_int_to_bts
    )


def get_property_translator(
        varname,
        varname_regex,
        cb_get_description=lambda value, value_name_dct: value_name_dct[value]):
    """
    Get a descriptor allowing to make a value->name translation for variable values

    varname_regex -- The regex to find variables.
    cb_get_description -- lambda value, value_name_dct: Description
    return -- property allowing get-access to get an descriptive name
    """
    # Get globals of calling module containing the variables in question
    globals_caller = inspect.stack()[1][0].f_globals

    def collect_cb():
        varname_pattern = re.compile(varname_regex)
        return {value: name for name, value in globals_caller.items() if
                type(value) in VARFILTER_TYPES and varname_pattern.match(name)}

    ldict = LazyDict(collect_cb)

    def descricptor_cb(value):
        if value not in ldict:
            return ""

        # logger.debug("Trying to get description!")
        return cb_get_description(value, ldict)

    # Only get access
    return property(
        lambda obj: descricptor_cb(obj.__getattribute__(varname))
    )


def get_ondemand_property(varname, initval_cb):
    """
    Creates a property whose value gets initialized ondemand.
    This is meant as an alternative to an initialization in __init__
    to decrease initial loading time
    """
    varname_shadowed = "_%s" % varname

    def get_var(self):
        try:
            return self.__getattribute__(varname_shadowed)
        except:
            val = initval_cb()
            self.__setattr__(varname_shadowed, val)
            return val

    def set_var(self, value):
        return self.__setattr__(varname_shadowed, value)

    return property(get_var, set_var)
